# Visualisation tools
#
# Copyright (C) Andrew Bartlett 2015, 2018
#
# by Douglas Bagnall <douglas.bagnall@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from collections import defaultdict
import subprocess
import tempfile
import samba.getopt as options
from samba import dsdb
from samba import nttime2unix
from samba.netcmd import Command, SuperCommand, CommandError, Option
from samba.samdb import SamDB
from samba.graph import dot_graph
from samba.graph import distance_matrix, COLOUR_SETS
from samba.graph import full_matrix
from samba.colour import is_colour_wanted

from ldb import SCOPE_BASE, SCOPE_SUBTREE, LdbError
import time
import re
from samba.kcc import KCC, ldif_import_export
from samba.kcc.kcc_utils import KCCError
from samba.uptodateness import (
    get_partition_maps,
    get_partition,
    get_own_cursor,
    get_utdv,
    get_utdv_edges,
    get_utdv_distances,
    get_utdv_max_distance,
    get_kcc_and_dsas,
)

COMMON_OPTIONS = [
    Option("-H", "--URL", help="LDB URL for database or target server",
           type=str, metavar="URL", dest="H"),
    Option("-o", "--output", help="write here (default stdout)",
           type=str, metavar="FILE", default=None),
    Option("--distance", help="Distance matrix graph output (default)",
           dest='format', const='distance', action='store_const'),
    Option("--utf8", help="Use utf-8 Unicode characters",
           action='store_true'),
    Option("--color-scheme", help=("use this colour scheme "
                                   "(implies --color=yes)"),
           choices=list(COLOUR_SETS.keys())),
    Option("-S", "--shorten-names",
           help="don't print long common suffixes",
           action='store_true', default=False),
    Option("-r", "--talk-to-remote", help="query other DCs' databases",
           action='store_true', default=False),
    Option("--no-key", help="omit the explanatory key",
           action='store_false', default=True, dest='key'),
]

DOT_OPTIONS = [
    Option("--dot", help="Graphviz dot output", dest='format',
           const='dot', action='store_const'),
    Option("--xdot", help="attempt to call Graphviz xdot", dest='format',
           const='xdot', action='store_const'),
]

TEMP_FILE = '__temp__'


class GraphCommand(Command):
    """Base class for graphing commands"""

    synopsis = "%prog [options]"
    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }
    takes_options = COMMON_OPTIONS + DOT_OPTIONS
    takes_args = ()

    def get_db(self, H, sambaopts, credopts):
        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp, fallback_machine=True)
        samdb = SamDB(url=H, credentials=creds, lp=lp)
        return samdb

    def write(self, s, fn=None, suffix='.dot'):
        """Decide whether we're dealing with a filename, a tempfile, or
        stdout, and write accordingly.

        :param s: the string to write
        :param fn: a destination
        :param suffix: suffix, if destination is a tempfile

        If fn is None or "-", write to stdout.
        If fn is visualize.TEMP_FILE, write to a temporary file
        Otherwise fn should be a filename to write to.
        """
        if fn is None or fn == '-':
            # we're just using stdout (a.k.a self.outf)
            print(s, file=self.outf)
            return

        if fn is TEMP_FILE:
            fd, fn = tempfile.mkstemp(prefix='samba-tool-visualise',
                                      suffix=suffix)
            f = open(fn, 'w')
            os.close(fd)
        else:
            f = open(fn, 'w')

        f.write(s)
        f.close()
        return fn

    def calc_output_format(self, format, output):
        """Heuristics to work out what output format was wanted."""
        if not format:
            # They told us nothing! We have to work it out for ourselves.
            if output and output.lower().endswith('.dot'):
                return 'dot'
            else:
                return 'distance'

        if format == 'xdot':
            return 'dot'

        return format

    def call_xdot(self, s, output):
        if output is None:
            fn = self.write(s, TEMP_FILE)
        else:
            fn = self.write(s, output)
        xdot = os.environ.get('SAMBA_TOOL_XDOT_PATH', '/usr/bin/xdot')
        subprocess.call([xdot, fn])
        os.remove(fn)

    def calc_distance_color_scheme(self, color_scheme, output):
        """Heuristics to work out the colour scheme for distance matrices.
        Returning None means no colour, otherwise it should be a colour
        from graph.COLOUR_SETS"""
        if color_scheme is not None:
            # --color-scheme implies --color=yes for *this* purpose.
            return color_scheme

        if output in ('-', None):
            output = self.outf

        want_colour = is_colour_wanted(output, hint=self.requested_colour)
        if not want_colour:
            return None

        # if we got to here, we are using colour according to the
        # --color/NO_COLOR rules, but no colour scheme has been
        # specified, so we choose some defaults.
        if '256color' in os.environ.get('TERM', ''):
            return 'xterm-256color-heatmap'
        return 'ansi'


def get_dnstr_site(dn):
    """Helper function for sorting and grouping DNs by site, if
    possible."""
    m = re.search(r'CN=Servers,CN=\s*([^,]+)\s*,CN=Sites', dn)
    if m:
        return m.group(1)
    # Oh well, let it sort by DN
    return dn


def get_dnstrlist_site(t):
    """Helper function for sorting and grouping lists of (DN, ...) tuples
    by site, if possible."""
    return get_dnstr_site(t[0])


def colour_hash(x):
    """Generate a randomish but consistent darkish colour based on the
    given object."""
    from hashlib import md5
    tmp_str = str(x)
    if isinstance(tmp_str, str):
        tmp_str = tmp_str.encode('utf8')
    c = int(md5(tmp_str).hexdigest()[:6], base=16) & 0x7f7f7f
    return '#%06x' % c


class cmd_reps(GraphCommand):
    "repsFrom/repsTo from every DSA"

    takes_options = COMMON_OPTIONS + DOT_OPTIONS + [
        Option("-p", "--partition", help="restrict to this partition",
               default=None),
    ]

    def run(self, H=None, output=None, shorten_names=False,
            key=True, talk_to_remote=False,
            sambaopts=None, credopts=None, versionopts=None,
            mode='self', partition=None, color_scheme=None,
            utf8=None, format=None, xdot=False):
        # We use the KCC libraries in readonly mode to get the
        # replication graph.
        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp, fallback_machine=True)
        local_kcc, dsas = get_kcc_and_dsas(H, lp, creds)
        unix_now = local_kcc.unix_now

        partition = get_partition(local_kcc.samdb, partition)

        # nc_reps is an autovivifying dictionary of dictionaries of lists.
        # nc_reps[partition]['current' | 'needed'] is a list of
        # (dsa dn string, repsFromTo object) pairs.
        nc_reps = defaultdict(lambda: defaultdict(list))

        guid_to_dnstr = {}

        # We run a new KCC for each DSA even if we aren't talking to
        # the remote, because after kcc.run (or kcc.list_dsas) the kcc
        # ends up in a messy state.
        for dsa_dn in dsas:
            kcc = KCC(unix_now, readonly=True)
            if talk_to_remote:
                res = local_kcc.samdb.search(dsa_dn,
                                             scope=SCOPE_BASE,
                                             attrs=["dNSHostName"])
                dns_name = str(res[0]["dNSHostName"][0])
                print("Attempting to contact ldap://%s (%s)" %
                      (dns_name, dsa_dn),
                      file=sys.stderr)
                try:
                    kcc.load_samdb("ldap://%s" % dns_name, lp, creds)
                except KCCError as e:
                    print("Could not contact ldap://%s (%s)" % (dns_name, e),
                          file=sys.stderr)
                    continue

                kcc.run(H, lp, creds)
            else:
                kcc.load_samdb(H, lp, creds)
                kcc.run(H, lp, creds, forced_local_dsa=dsa_dn)

            dsas_from_here = set(kcc.list_dsas())
            if dsas != dsas_from_here:
                print("found extra DSAs:", file=sys.stderr)
                for dsa in (dsas_from_here - dsas):
                    print("   %s" % dsa, file=sys.stderr)
                print("missing DSAs (known locally, not by %s):" % dsa_dn,
                      file=sys.stderr)
                for dsa in (dsas - dsas_from_here):
                    print("   %s" % dsa, file=sys.stderr)

            for remote_dn in dsas_from_here:
                if mode == 'others' and remote_dn == dsa_dn:
                    continue
                elif mode == 'self' and remote_dn != dsa_dn:
                    continue

                remote_dsa = kcc.get_dsa('CN=NTDS Settings,' + remote_dn)
                kcc.translate_ntdsconn(remote_dsa)
                guid_to_dnstr[str(remote_dsa.dsa_guid)] = remote_dn
                # get_reps_tables() returns two dictionaries mapping
                # dns to NCReplica objects
                c, n = remote_dsa.get_rep_tables()
                for part, rep in c.items():
                    if partition is None or part == partition:
                        nc_reps[part]['current'].append((dsa_dn, rep))
                for part, rep in n.items():
                    if partition is None or part == partition:
                        nc_reps[part]['needed'].append((dsa_dn, rep))

        all_edges = {'needed': {'to': [], 'from': []},
                     'current': {'to': [], 'from': []}}

        short_partitions, long_partitions = get_partition_maps(local_kcc.samdb)

        for partname, part in nc_reps.items():
            for state, edgelists in all_edges.items():
                for dsa_dn, rep in part[state]:
                    short_name = long_partitions.get(partname, partname)
                    for r in rep.rep_repsFrom:
                        edgelists['from'].append(
                            (dsa_dn,
                             guid_to_dnstr[str(r.source_dsa_obj_guid)],
                             short_name))
                    for r in rep.rep_repsTo:
                        edgelists['to'].append(
                            (guid_to_dnstr[str(r.source_dsa_obj_guid)],
                             dsa_dn,
                             short_name))

        # Here we have the set of edges. From now it is a matter of
        # interpretation and presentation.

        if self.calc_output_format(format, output) == 'distance':
            color_scheme = self.calc_distance_color_scheme(color_scheme,
                                                           output)
            header_strings = {
                'from': "RepsFrom objects for %s",
                'to': "RepsTo objects for %s",
            }
            for state, edgelists in all_edges.items():
                for direction, items in edgelists.items():
                    part_edges = defaultdict(list)
                    for src, dest, part in items:
                        part_edges[part].append((src, dest))
                    for part, edges in part_edges.items():
                        s = distance_matrix(None, edges,
                                            utf8=utf8,
                                            colour=color_scheme,
                                            shorten_names=shorten_names,
                                            generate_key=key,
                                            grouping_function=get_dnstr_site)

                        s = "\n%s\n%s" % (header_strings[direction] % part, s)
                        self.write(s, output)
            return

        edge_colours = []
        edge_styles = []
        dot_edges = []
        dot_vertices = set()
        used_colours = {}
        key_set = set()
        for state, edgelist in all_edges.items():
            for direction, items in edgelist.items():
                for src, dest, part in items:
                    colour = used_colours.setdefault((part),
                                                     colour_hash((part,
                                                                  direction)))
                    linestyle = 'dotted' if state == 'needed' else 'solid'
                    arrow = 'open' if direction == 'to' else 'empty'
                    dot_vertices.add(src)
                    dot_vertices.add(dest)
                    dot_edges.append((src, dest))
                    edge_colours.append(colour)
                    style = 'style="%s"; arrowhead=%s' % (linestyle, arrow)
                    edge_styles.append(style)
                    key_set.add((part, 'reps' + direction.title(),
                                 colour, style))

        key_items = []
        if key:
            for part, direction, colour, linestyle in sorted(key_set):
                key_items.append((False,
                                  'color="%s"; %s' % (colour, linestyle),
                                  "%s %s" % (part, direction)))
            key_items.append((False,
                              'style="dotted"; arrowhead="open"',
                              "repsFromTo is needed"))
            key_items.append((False,
                              'style="solid"; arrowhead="open"',
                              "repsFromTo currently exists"))

        s = dot_graph(dot_vertices, dot_edges,
                      directed=True,
                      edge_colors=edge_colours,
                      edge_styles=edge_styles,
                      shorten_names=shorten_names,
                      key_items=key_items)

        if format == 'xdot':
            self.call_xdot(s, output)
        else:
            self.write(s, output)


class NTDSConn(object):
    """Collects observation counts for NTDS connections, so we know
    whether all DSAs agree."""
    def __init__(self, src, dest):
        self.observations = 0
        self.src_attests = False
        self.dest_attests = False
        self.src = src
        self.dest = dest

    def attest(self, attester):
        self.observations += 1
        if attester == self.src:
            self.src_attests = True
        if attester == self.dest:
            self.dest_attests = True


class cmd_ntdsconn(GraphCommand):
    "Draw the NTDSConnection graph"
    takes_options = COMMON_OPTIONS + DOT_OPTIONS + [
        Option("--importldif", help="graph from samba_kcc generated ldif",
               default=None),
    ]

    def import_ldif_db(self, ldif, lp):
        d = tempfile.mkdtemp(prefix='samba-tool-visualise')
        fn = os.path.join(d, 'imported.ldb')
        self._tmp_fn_to_delete = fn
        samdb = ldif_import_export.ldif_to_samdb(fn, lp, ldif)
        return fn

    def run(self, H=None, output=None, shorten_names=False,
            key=True, talk_to_remote=False,
            sambaopts=None, credopts=None, versionopts=None,
            color_scheme=None,
            utf8=None, format=None, importldif=None,
            xdot=False):

        lp = sambaopts.get_loadparm()
        if importldif is None:
            creds = credopts.get_credentials(lp, fallback_machine=True)
        else:
            creds = None
            H = self.import_ldif_db(importldif, lp)

        local_kcc, dsas = get_kcc_and_dsas(H, lp, creds)
        local_dsa_dn = local_kcc.my_dsa_dnstr.split(',', 1)[1]
        vertices = set()
        attested_edges = []
        for dsa_dn in dsas:
            if talk_to_remote:
                res = local_kcc.samdb.search(dsa_dn,
                                             scope=SCOPE_BASE,
                                             attrs=["dNSHostName"])
                dns_name = res[0]["dNSHostName"][0]
                try:
                    samdb = self.get_db("ldap://%s" % dns_name, sambaopts,
                                        credopts)
                except LdbError as e:
                    print("Could not contact ldap://%s (%s)" % (dns_name, e),
                          file=sys.stderr)
                    continue

                ntds_dn = samdb.get_dsServiceName()
                dn = samdb.domain_dn()
            else:
                samdb = self.get_db(H, sambaopts, credopts)
                ntds_dn = 'CN=NTDS Settings,' + dsa_dn
                dn = dsa_dn

            res = samdb.search(ntds_dn,
                               scope=SCOPE_BASE,
                               attrs=["msDS-isRODC"])

            is_rodc = res[0]["msDS-isRODC"][0] == 'TRUE'

            vertices.add((ntds_dn, 'RODC' if is_rodc else ''))
            # XXX we could also look at schedule
            res = samdb.search(dn,
                               scope=SCOPE_SUBTREE,
                               expression="(objectClass=nTDSConnection)",
                               attrs=['fromServer'],
                               # XXX can't be critical for ldif test
                               # controls=["search_options:1:2"],
                               controls=["search_options:0:2"],
                               )

            for msg in res:
                msgdn = str(msg.dn)
                dest_dn = msgdn[msgdn.index(',') + 1:]
                attested_edges.append((str(msg['fromServer'][0]),
                                       dest_dn, ntds_dn))

        if importldif and H == self._tmp_fn_to_delete:
            os.remove(H)
            os.rmdir(os.path.dirname(H))

        # now we overlay all the graphs and generate styles accordingly
        edges = {}
        for src, dest, attester in attested_edges:
            k = (src, dest)
            if k in edges:
                e = edges[k]
            else:
                e = NTDSConn(*k)
                edges[k] = e
            e.attest(attester)

        vertices, rodc_status = zip(*sorted(vertices))

        if self.calc_output_format(format, output) == 'distance':
            color_scheme = self.calc_distance_color_scheme(color_scheme,
                                                           output)
            colours = COLOUR_SETS[color_scheme]
            c_header = colours.get('header', '')
            c_reset = colours.get('reset', '')

            epilog = []
            if 'RODC' in rodc_status:
                epilog.append('No outbound connections are expected from RODCs')

            if not talk_to_remote:
                # If we are not talking to remote servers, we list all
                # the connections.
                graph_edges = edges.keys()
                title = 'NTDS Connections known to %s' % local_dsa_dn

            else:
                # If we are talking to the remotes, there are
                # interesting cases we can discover. What matters most
                # is that the destination (i.e. owner) knowns about
                # the connection, but it would be worth noting if the
                # source doesn't. Another strange situation could be
                # when a DC thinks there is a connection elsewhere,
                # but the computers allegedly involved don't believe
                # it exists.
                #
                # With limited bandwidth in the table, we mark the
                # edges known to the destination, and note the other
                # cases in a list after the diagram.
                graph_edges = []
                source_denies = []
                dest_denies = []
                both_deny = []
                for e, conn in edges.items():
                    if conn.dest_attests:
                        graph_edges.append(e)
                        if not conn.src_attests:
                            source_denies.append(e)
                    elif conn.src_attests:
                        dest_denies.append(e)
                    else:
                        both_deny.append(e)

                title = 'NTDS Connections known to each destination DC'

                if both_deny:
                    epilog.append('The following connections are alleged by '
                                  'DCs other than the source and '
                                  'destination:\n')
                    for e in both_deny:
                        epilog.append('  %s -> %s\n' % e)
                if dest_denies:
                    epilog.append('The following connections are alleged by '
                                  'DCs other than the destination but '
                                  'including the source:\n')
                    for e in dest_denies:
                        epilog.append('  %s -> %s\n' % e)
                if source_denies:
                    epilog.append('The following connections '
                                  '(included in the chart) '
                                  'are not known to the source DC:\n')
                    for e in source_denies:
                        epilog.append('  %s -> %s\n' % e)

            s = distance_matrix(vertices, graph_edges,
                                utf8=utf8,
                                colour=color_scheme,
                                shorten_names=shorten_names,
                                generate_key=key,
                                grouping_function=get_dnstrlist_site,
                                row_comments=rodc_status)

            epilog = ''.join(epilog)
            if epilog:
                epilog = '\n%sNOTES%s\n%s' % (c_header,
                                              c_reset,
                                              epilog)

            self.write('\n%s\n\n%s\n%s' % (title,
                                           s,
                                           epilog), output)
            return

        dot_edges = []
        edge_colours = []
        edge_styles = []
        edge_labels = []
        n_servers = len(dsas)
        for k, e in sorted(edges.items()):
            dot_edges.append(k)
            if e.observations == n_servers or not talk_to_remote:
                edge_colours.append('#000000')
                edge_styles.append('')
            elif e.dest_attests:
                edge_styles.append('')
                if e.src_attests:
                    edge_colours.append('#0000ff')
                else:
                    edge_colours.append('#cc00ff')
            elif e.src_attests:
                edge_colours.append('#ff0000')
                edge_styles.append('style=dashed')
            else:
                edge_colours.append('#ff0000')
                edge_styles.append('style=dotted')

        key_items = []
        if key:
            key_items.append((False,
                              'color="#000000"',
                              "NTDS Connection"))
            for colour, desc in (('#0000ff', "missing from some DCs"),
                                 ('#cc00ff', "missing from source DC")):
                if colour in edge_colours:
                    key_items.append((False, 'color="%s"' % colour, desc))

            for style, desc in (('style=dashed', "unknown to destination"),
                                ('style=dotted',
                                 "unknown to source and destination")):
                if style in edge_styles:
                    key_items.append((False,
                                      'color="#ff0000; %s"' % style,
                                      desc))

        if talk_to_remote:
            title = 'NTDS Connections'
        else:
            title = 'NTDS Connections known to %s' % local_dsa_dn

        s = dot_graph(sorted(vertices), dot_edges,
                      directed=True,
                      title=title,
                      edge_colors=edge_colours,
                      edge_labels=edge_labels,
                      edge_styles=edge_styles,
                      shorten_names=shorten_names,
                      key_items=key_items)

        if format == 'xdot':
            self.call_xdot(s, output)
        else:
            self.write(s, output)


class cmd_uptodateness(GraphCommand):
    """visualize uptodateness vectors"""

    takes_options = COMMON_OPTIONS + [
        Option("-p", "--partition", help="restrict to this partition",
               default=None),
        Option("--max-digits", default=3, type=int,
               help="display this many digits of out-of-date-ness"),
    ]

    def run(self, H=None, output=None, shorten_names=False,
            key=True, talk_to_remote=False,
            sambaopts=None, credopts=None, versionopts=None,
            color_scheme=None,
            utf8=False, format=None, importldif=None,
            xdot=False, partition=None, max_digits=3):
        if not talk_to_remote:
            print("this won't work without talking to the remote servers "
                  "(use -r)", file=self.outf)
            return

        # We use the KCC libraries in readonly mode to get the
        # replication graph.
        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp, fallback_machine=True)
        local_kcc, dsas = get_kcc_and_dsas(H, lp, creds)
        self.samdb = local_kcc.samdb
        partition = get_partition(self.samdb, partition)

        short_partitions, long_partitions = get_partition_maps(self.samdb)
        color_scheme = self.calc_distance_color_scheme(color_scheme,
                                                       output)

        for part_name, part_dn in short_partitions.items():
            if partition not in (part_dn, None):
                continue  # we aren't doing this partition

            utdv_edges = get_utdv_edges(local_kcc, dsas, part_dn, lp, creds)

            distances = get_utdv_distances(utdv_edges, dsas)

            max_distance = get_utdv_max_distance(distances)

            digits = min(max_digits, len(str(max_distance)))
            if digits < 1:
                digits = 1
            c_scale = 10 ** digits

            s = full_matrix(distances,
                            utf8=utf8,
                            colour=color_scheme,
                            shorten_names=shorten_names,
                            generate_key=key,
                            grouping_function=get_dnstr_site,
                            colour_scale=c_scale,
                            digits=digits,
                            ylabel='DC',
                            xlabel='out-of-date-ness')

            self.write('\n%s\n\n%s' % (part_name, s), output)


class cmd_visualize(SuperCommand):
    """Produces graphical representations of Samba network state."""
    subcommands = {}

    for k, v in globals().items():
        if k.startswith('cmd_'):
            subcommands[k[4:]] = v()



# python site manipulation code
# Copyright Matthieu Patou <mat@matws.net> 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Manipulating sites."""

import ldb
from ldb import FLAG_MOD_ADD, LdbError


class SiteException(Exception):
    """Base element for Sites errors"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, self.value)


class SiteNotFoundException(SiteException):
    """Raised when the site is not found and it's expected to exists."""


class SiteAlreadyExistsException(SiteException):
    """Raised when the site is not found and it's expected not to exists."""


class SiteServerNotEmptyException(SiteException):
    """Raised when the site still has servers attached."""


def create_site(samdb, configDn, siteName):
    """
    Create a site

    :param samdb: A samdb connection
    :param configDn: The DN of the configuration partition
    :param siteName: Name of the site to create
    :return: True upon success
    :raise SiteAlreadyExists: if the site to be created already exists.
    """

    ret = samdb.search(base=configDn, scope=ldb.SCOPE_SUBTREE,
                       expression='(&(objectclass=Site)(cn=%s))' % siteName)
    if len(ret) != 0:
        raise SiteAlreadyExistsException('A site with the name %s already exists' % siteName)

    m = ldb.Message()
    m.dn = ldb.Dn(samdb, "Cn=%s,CN=Sites,%s" % (siteName, str(configDn)))
    m["objectclass"] = ldb.MessageElement("site", FLAG_MOD_ADD, "objectclass")

    samdb.add(m)

    m2 = ldb.Message()
    m2.dn = ldb.Dn(samdb, "Cn=NTDS Site Settings,%s" % str(m.dn))
    m2["objectclass"] = ldb.MessageElement("nTDSSiteSettings", FLAG_MOD_ADD, "objectclass")

    samdb.add(m2)

    m3 = ldb.Message()
    m3.dn = ldb.Dn(samdb, "Cn=Servers,%s" % str(m.dn))
    m3["objectclass"] = ldb.MessageElement("serversContainer", FLAG_MOD_ADD, "objectclass")

    samdb.add(m3)

    return True


def delete_site(samdb, configDn, siteName):
    """
    Delete a site

    :param samdb: A samdb connection
    :param configDn: The DN of the configuration partition
    :param siteName: Name of the site to delete
    :return: True upon success
    :raise SiteNotFoundException: if the site to be deleted do not exists.
    :raise SiteServerNotEmpty: if the site has still servers in it.
    """

    dnsite = ldb.Dn(samdb, "CN=Sites")
    try:
        dnsite.add_base(configDn)
    except ldb.LdbError:
        raise SiteException("dnsite.add_base() failed")
    try:
        dnsite.add_child("CN=X")
    except ldb.LdbError:
        raise SiteException("dnsite.add_child() failed")
    dnsite.set_component(0, "CN", siteName)

    dnservers = ldb.Dn(samdb, "CN=Servers")
    dnservers.add_base(dnsite)

    try:
        ret = samdb.search(base=dnsite, scope=ldb.SCOPE_BASE,
                           expression="objectClass=site")
        if len(ret) != 1:
            raise SiteNotFoundException('Site %s does not exist' % siteName)
    except LdbError as e:
        (enum, estr) = e.args
        if enum == ldb.ERR_NO_SUCH_OBJECT:
            raise SiteNotFoundException('Site %s does not exist' % siteName)

    ret = samdb.search(base=dnservers, scope=ldb.SCOPE_ONELEVEL,
                       expression='(objectclass=server)')
    if len(ret) != 0:
        raise SiteServerNotEmptyException('Site %s still has servers in it, move them before removal' % siteName)

    samdb.delete(dnsite, ["tree_delete:0"])

    return True



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2005-2008
# Copyright (C) Martin Kuehl <mkhl@samba.org> 2006
#
# This is a Python port of the original in testprogs/ejs/samba3sam.js
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for the samba3sam LDB module, which maps Samba3 LDAP to AD LDAP."""

import os
import ldb
from ldb import SCOPE_DEFAULT, SCOPE_BASE
from samba import Ldb, substitute_var
from samba.tests import TestCaseInTempDir, env_loadparm
import samba.dcerpc.security
import samba.ndr
from samba.auth import system_session
from operator import attrgetter


def read_datafile(filename):
    paths = ["../../../../../testdata/samba3",
             "../../../../testdata/samba3"]
    for p in paths:
        datadir = os.path.join(os.path.dirname(__file__), p)
        if os.path.exists(datadir):
            break
    return open(os.path.join(datadir, filename), 'r').read()


def ldb_debug(l, text):
    print(text)


class MapBaseTestCase(TestCaseInTempDir):
    """Base test case for mapping tests."""

    def setup_modules(self, ldb, s3, s4):
        ldb.add({"dn": "@MAP=samba3sam",
                 "@FROM": s4.basedn,
                 "@TO": "sambaDomainName=TESTS," + s3.basedn})

        ldb.add({"dn": "@MODULES",
                 "@LIST": "rootdse,dsdb_paged_results,server_sort,asq,samldb,password_hash,operational,objectguid,rdn_name,samba3sam,samba3sid,show_deleted_ignore,dsdb_flags_ignore,partition"})

        ldb.add({"dn": "@PARTITION",
                 "partition": ["%s" % (s4.basedn_casefold),
                               "%s" % (s3.basedn_casefold)],
                 "replicateEntries": ["@ATTRIBUTES", "@INDEXLIST"],
                 "modules": "*:"})

    def setUp(self):
        self.lp = env_loadparm()
        self.lp.set("workgroup", "TESTS")
        self.lp.set("netbios name", "TESTS")
        super().setUp()

        def make_dn(basedn, rdn):
            return "%s,sambaDomainName=TESTS,%s" % (rdn, basedn)

        def make_s4dn(basedn, rdn):
            return "%s,%s" % (rdn, basedn)

        self.ldbfile = os.path.join(self.tempdir, "sam.ldb")
        self.ldburl = "tdb://" + self.ldbfile

        tempdir = self.tempdir

        class Target:
            """Simple helper class that contains data for a specific SAM
            connection."""

            def __init__(self, basedn, dn, lp):
                self.db = Ldb(lp=lp, session_info=system_session())
                self.db.set_opaque("skip_allocate_sids", "true")
                self.basedn = basedn
                self.basedn_casefold = ldb.Dn(self.db, basedn).get_casefold()
                self.substvars = {"BASEDN": self.basedn}
                self.file = os.path.join(tempdir, "%s.ldb" % self.basedn_casefold)
                self.url = "tdb://" + self.file
                self._dn = dn

            def dn(self, rdn):
                return self._dn(self.basedn, rdn)

            def connect(self):
                return self.db.connect(self.url)

            def setup_data(self, path):
                self.add_ldif(read_datafile(path))

            def subst(self, text):
                return substitute_var(text, self.substvars)

            def add_ldif(self, ldif):
                self.db.add_ldif(self.subst(ldif))

            def modify_ldif(self, ldif):
                self.db.modify_ldif(self.subst(ldif))

        self.samba4 = Target("dc=vernstok,dc=nl", make_s4dn, self.lp)
        self.samba3 = Target("cn=Samba3Sam", make_dn, self.lp)

        self.samba3.connect()
        self.samba4.connect()

    def tearDown(self):
        os.unlink(self.ldbfile)
        os.unlink(self.samba3.file)
        os.unlink(self.samba4.file)
        pdir = "%s.d" % self.ldbfile
        mdata = os.path.join(pdir, "metadata.tdb")
        if os.path.exists(mdata):
            os.unlink(mdata)
            os.rmdir(pdir)
        super().tearDown()

    def assertSidEquals(self, text, ndr_sid):
        sid_obj1 = samba.ndr.ndr_unpack(samba.dcerpc.security.dom_sid,
                                        ndr_sid[0])
        sid_obj2 = samba.dcerpc.security.dom_sid(text)
        self.assertEqual(sid_obj1, sid_obj2)


class Samba3SamTestCase(MapBaseTestCase):

    def setUp(self):
        super().setUp()
        ldb = Ldb(self.ldburl, lp=self.lp, session_info=system_session())
        ldb.set_opaque("skip_allocate_sids", "true")
        self.samba3.setup_data("samba3.ldif")
        ldif = read_datafile("provision_samba3sam.ldif")
        ldb.add_ldif(self.samba4.subst(ldif))
        self.setup_modules(ldb, self.samba3, self.samba4)
        del ldb
        self.ldb = Ldb(self.ldburl, lp=self.lp, session_info=system_session())
        self.ldb.set_opaque("skip_allocate_sids", "true")

    def test_search_non_mapped(self):
        """Looking up by non-mapped attribute"""
        msg = self.ldb.search(expression="(cn=Administrator)")
        self.assertEqual(len(msg), 1)
        self.assertEqual(str(msg[0]["cn"]), "Administrator")

    def test_search_mapped(self):
        """Looking up by mapped attribute"""
        msg = self.ldb.search(expression="(name=Backup Operators)")
        self.assertEqual(len(msg), 1)
        self.assertEqual(str(msg[0]["name"]), "Backup Operators")

    def test_old_name_of_renamed(self):
        """Looking up by old name of renamed attribute"""
        msg = self.ldb.search(expression="(displayName=Backup Operators)")
        self.assertEqual(len(msg), 0)

    def test_mapped_containing_sid(self):
        """Looking up mapped entry containing SID"""
        msg = self.ldb.search(expression="(cn=Replicator)")
        self.assertEqual(len(msg), 1)
        self.assertEqual(str(msg[0].dn),
                          "cn=Replicator,ou=Groups,dc=vernstok,dc=nl")
        self.assertTrue("objectSid" in msg[0])
        self.assertSidEquals("S-1-5-21-4231626423-2410014848-2360679739-1052",
                             msg[0]["objectSid"])
        oc = set(msg[0]["objectClass"])
        self.assertEqual(oc, set([b"group"]))

    def test_search_by_objclass(self):
        """Looking up by objectClass"""
        msg = self.ldb.search(expression="(|(objectClass=user)(cn=Administrator))")
        self.assertEqual(set([str(m.dn) for m in msg]),
                          set(["unixName=Administrator,ou=Users,dc=vernstok,dc=nl",
                               "unixName=nobody,ou=Users,dc=vernstok,dc=nl"]))

    def test_s3sam_modify(self):
        # Adding a record that will be fallbacked
        self.ldb.add({
            "dn": "cn=Foo",
            "foo": "bar",
            "blah": "Blie",
            "cn": "Foo",
            "showInAdvancedViewOnly": "TRUE"})

        # Checking for existence of record (local)
        # TODO: This record must be searched in the local database, which is
        # currently only supported for base searches
        # msg = ldb.search(expression="(cn=Foo)", ['foo','blah','cn','showInAdvancedViewOnly')]
        # TODO: Actually, this version should work as well but doesn't...
        #
        #
        msg = self.ldb.search(expression="(cn=Foo)", base="cn=Foo",
                              scope=SCOPE_BASE,
                              attrs=['foo', 'blah', 'cn', 'showInAdvancedViewOnly'])
        self.assertEqual(len(msg), 1)
        self.assertEqual(str(msg[0]["showInAdvancedViewOnly"]), "TRUE")
        self.assertEqual(str(msg[0]["foo"]), "bar")
        self.assertEqual(str(msg[0]["blah"]), "Blie")

        # Adding record that will be mapped
        self.ldb.add({"dn": "cn=Niemand,cn=Users,dc=vernstok,dc=nl",
                      "objectClass": "user",
                      "unixName": "bin",
                      "sambaUnicodePwd": "geheim",
                      "cn": "Niemand"})

        # Checking for existence of record (remote)
        msg = self.ldb.search(expression="(unixName=bin)",
                              attrs=['unixName', 'cn', 'dn', 'sambaUnicodePwd'])
        self.assertEqual(len(msg), 1)
        self.assertEqual(str(msg[0]["cn"]), "Niemand")
        self.assertEqual(str(msg[0]["sambaUnicodePwd"]), "geheim")

        # Checking for existence of record (local && remote)
        msg = self.ldb.search(expression="(&(unixName=bin)(sambaUnicodePwd=geheim))",
                              attrs=['unixName', 'cn', 'dn', 'sambaUnicodePwd'])
        self.assertEqual(len(msg), 1)           # TODO: should check with more records
        self.assertEqual(str(msg[0]["cn"]), "Niemand")
        self.assertEqual(str(msg[0]["unixName"]), "bin")
        self.assertEqual(str(msg[0]["sambaUnicodePwd"]), "geheim")

        # Checking for existence of record (local || remote)
        msg = self.ldb.search(expression="(|(unixName=bin)(sambaUnicodePwd=geheim))",
                              attrs=['unixName', 'cn', 'dn', 'sambaUnicodePwd'])
        # print "got %d replies" % len(msg)
        self.assertEqual(len(msg), 1)        # TODO: should check with more records
        self.assertEqual(str(msg[0]["cn"]), "Niemand")
        self.assertEqual(str(msg[0]["unixName"]), "bin")
        self.assertEqual(str(msg[0]["sambaUnicodePwd"]), "geheim")

        # Checking for data in destination database
        msg = self.samba3.db.search(expression="(cn=Niemand)")
        self.assertTrue(len(msg) >= 1)
        self.assertEqual(str(msg[0]["sambaSID"]),
                          "S-1-5-21-4231626423-2410014848-2360679739-2001")
        self.assertEqual(str(msg[0]["displayName"]), "Niemand")

        # Adding attribute...
        self.ldb.modify_ldif("""
dn: cn=Niemand,cn=Users,dc=vernstok,dc=nl
changetype: modify
add: description
description: Blah
""")

        # Checking whether changes are still there...
        msg = self.ldb.search(expression="(cn=Niemand)")
        self.assertTrue(len(msg) >= 1)
        self.assertEqual(str(msg[0]["cn"]), "Niemand")
        self.assertEqual(str(msg[0]["description"]), "Blah")

        # Modifying attribute...
        self.ldb.modify_ldif("""
dn: cn=Niemand,cn=Users,dc=vernstok,dc=nl
changetype: modify
replace: description
description: Blie
""")

        # Checking whether changes are still there...
        msg = self.ldb.search(expression="(cn=Niemand)")
        self.assertTrue(len(msg) >= 1)
        self.assertEqual(str(msg[0]["description"]), "Blie")

        # Deleting attribute...
        self.ldb.modify_ldif("""
dn: cn=Niemand,cn=Users,dc=vernstok,dc=nl
changetype: modify
delete: description
""")

        # Checking whether changes are no longer there...
        msg = self.ldb.search(expression="(cn=Niemand)")
        self.assertTrue(len(msg) >= 1)
        self.assertTrue("description" not in msg[0])

        # Renaming record...
        self.ldb.rename("cn=Niemand,cn=Users,dc=vernstok,dc=nl",
                        "cn=Niemand2,cn=Users,dc=vernstok,dc=nl")

        # Checking whether DN has changed...
        msg = self.ldb.search(expression="(cn=Niemand2)")
        self.assertEqual(len(msg), 1)
        self.assertEqual(str(msg[0].dn),
                          "cn=Niemand2,cn=Users,dc=vernstok,dc=nl")

        # Deleting record...
        self.ldb.delete("cn=Niemand2,cn=Users,dc=vernstok,dc=nl")

        # Checking whether record is gone...
        msg = self.ldb.search(expression="(cn=Niemand2)")
        self.assertEqual(len(msg), 0)


class MapTestCase(MapBaseTestCase):

    def setUp(self):
        super().setUp()
        ldb = Ldb(self.ldburl, lp=self.lp, session_info=system_session())
        ldb.set_opaque("skip_allocate_sids", "true")
        ldif = read_datafile("provision_samba3sam.ldif")
        ldb.add_ldif(self.samba4.subst(ldif))
        self.setup_modules(ldb, self.samba3, self.samba4)
        del ldb
        self.ldb = Ldb(self.ldburl, lp=self.lp, session_info=system_session())
        self.ldb.set_opaque("skip_allocate_sids", "true")

    def test_map_search(self):
        """Running search tests on mapped data."""
        self.samba3.db.add({
            "dn": "sambaDomainName=TESTS," + self.samba3.basedn,
            "objectclass": ["sambaDomain", "top"],
            "sambaSID": "S-1-5-21-4231626423-2410014848-2360679739",
            "sambaNextRid": "2000",
            "sambaDomainName": "TESTS"
        })

        # Add a set of split records
        self.ldb.add_ldif("""
dn: """ + self.samba4.dn("cn=Domain Users") + """
objectClass: group
cn: Domain Users
objectSid: S-1-5-21-4231626423-2410014848-2360679739-513
""")

        # Add a set of split records
        self.ldb.add_ldif("""
dn: """ + self.samba4.dn("cn=X") + """
objectClass: user
cn: X
codePage: x
revision: x
dnsHostName: x
nextRid: y
lastLogon: x
description: x
objectSid: S-1-5-21-4231626423-2410014848-2360679739-1052
""")

        self.ldb.add({
            "dn": self.samba4.dn("cn=Y"),
            "objectClass": "top",
            "cn": "Y",
            "codePage": "x",
            "revision": "x",
            "dnsHostName": "y",
            "nextRid": "y",
            "lastLogon": "y",
            "description": "x"})

        self.ldb.add({
            "dn": self.samba4.dn("cn=Z"),
            "objectClass": "top",
            "cn": "Z",
            "codePage": "x",
            "revision": "y",
            "dnsHostName": "z",
            "nextRid": "y",
            "lastLogon": "z",
            "description": "y"})

        # Add a set of remote records

        self.samba3.db.add({
            "dn": self.samba3.dn("cn=A"),
            "objectClass": "posixAccount",
            "cn": "A",
            "sambaNextRid": "x",
            "sambaBadPasswordCount": "x",
            "sambaLogonTime": "x",
            "description": "x",
            "sambaSID": "S-1-5-21-4231626423-2410014848-2360679739-1052",
            "sambaPrimaryGroupSID": "S-1-5-21-4231626423-2410014848-2360679739-512"})

        self.samba3.db.add({
            "dn": self.samba3.dn("cn=B"),
            "objectClass": "top",
            "cn": "B",
            "sambaNextRid": "x",
            "sambaBadPasswordCount": "x",
            "sambaLogonTime": "y",
            "description": "x"})

        self.samba3.db.add({
            "dn": self.samba3.dn("cn=C"),
            "objectClass": "top",
            "cn": "C",
            "sambaNextRid": "x",
            "sambaBadPasswordCount": "y",
            "sambaLogonTime": "z",
            "description": "y"})

        # Testing search by DN

        # Search remote record by local DN
        dn = self.samba4.dn("cn=A")
        res = self.ldb.search(dn, scope=SCOPE_BASE,
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")

        # Search remote record by remote DN
        dn = self.samba3.dn("cn=A")
        res = self.samba3.db.search(dn, scope=SCOPE_BASE,
                                    attrs=["dnsHostName", "lastLogon", "sambaLogonTime"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("dnsHostName" not in res[0])
        self.assertTrue("lastLogon" not in res[0])
        self.assertEqual(str(res[0]["sambaLogonTime"]), "x")

        # Search split record by local DN
        dn = self.samba4.dn("cn=X")
        res = self.ldb.search(dn, scope=SCOPE_BASE,
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["dnsHostName"]), "x")
        self.assertEqual(str(res[0]["lastLogon"]), "x")

        # Search split record by remote DN
        dn = self.samba3.dn("cn=X")
        res = self.samba3.db.search(dn, scope=SCOPE_BASE,
                                    attrs=["dnsHostName", "lastLogon", "sambaLogonTime"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("dnsHostName" not in res[0])
        self.assertTrue("lastLogon" not in res[0])
        self.assertEqual(str(res[0]["sambaLogonTime"]), "x")

        # Testing search by attribute

        # Search by ignored attribute
        res = self.ldb.search(expression="(revision=x)", scope=SCOPE_DEFAULT,
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[0]["dnsHostName"]), "x")
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[1]["dnsHostName"]), "y")
        self.assertEqual(str(res[1]["lastLogon"]), "y")

        # Search by kept attribute
        res = self.ldb.search(expression="(description=y)",
                              scope=SCOPE_DEFAULT, attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "z")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[1]["dnsHostName"]), "z")
        self.assertEqual(str(res[1]["lastLogon"]), "z")

        # Search by renamed attribute
        res = self.ldb.search(expression="(badPwdCount=x)", scope=SCOPE_DEFAULT,
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")

        # Search by converted attribute
        # TODO:
        #   Using the SID directly in the parse tree leads to conversion
        #   errors, letting the search fail with no results.
        # res = self.ldb.search("(objectSid=S-1-5-21-4231626423-2410014848-2360679739-1052)", scope=SCOPE_DEFAULT, attrs)
        res = self.ldb.search(expression="(objectSid=*)", base=None, scope=SCOPE_DEFAULT, attrs=["dnsHostName", "lastLogon", "objectSid"])
        self.assertEqual(len(res), 4)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[1]["dnsHostName"]), "x")
        self.assertEqual(str(res[1]["lastLogon"]), "x")
        self.assertSidEquals("S-1-5-21-4231626423-2410014848-2360679739-1052",
                             res[1]["objectSid"])
        self.assertTrue("objectSid" in res[1])
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertSidEquals("S-1-5-21-4231626423-2410014848-2360679739-1052",
                             res[0]["objectSid"])
        self.assertTrue("objectSid" in res[0])

        # Search by generated attribute
        # In most cases, this even works when the mapping is missing
        # a `convert_operator' by enumerating the remote db.
        res = self.ldb.search(expression="(primaryGroupID=512)",
                              attrs=["dnsHostName", "lastLogon", "primaryGroupID"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[0]["primaryGroupID"]), "512")

        # Note that Xs "objectSid" seems to be fine in the previous search for
        # "objectSid"...
        # res = ldb.search(expression="(primaryGroupID=*)", NULL, ldb. SCOPE_DEFAULT, attrs)
        # print len(res) + " results found"
        # for i in range(len(res)):
        #    for (obj in res[i]) {
        #        print obj + ": " + res[i][obj]
        #    }
        #    print "---"
        #

        # Search by remote name of renamed attribute */
        res = self.ldb.search(expression="(sambaBadPasswordCount=*)",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 0)

        # Search by objectClass
        attrs = ["dnsHostName", "lastLogon", "objectClass"]
        res = self.ldb.search(expression="(objectClass=user)", attrs=attrs)
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[0]["objectClass"][0]), "user")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[1]["dnsHostName"]), "x")
        self.assertEqual(str(res[1]["lastLogon"]), "x")
        self.assertEqual(str(res[1]["objectClass"][0]), "user")

        # Prove that the objectClass is actually used for the search
        res = self.ldb.search(expression="(|(objectClass=user)(badPwdCount=x))",
                              attrs=attrs)
        self.assertEqual(len(res), 3)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[0]["objectClass"][0]), "user")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(set(res[1]["objectClass"]), set([b"top"]))
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[2]["dnsHostName"]), "x")
        self.assertEqual(str(res[2]["lastLogon"]), "x")
        self.assertEqual(str(res[2]["objectClass"][0]), "user")

        # Testing search by parse tree

        # Search by conjunction of local attributes
        res = self.ldb.search(expression="(&(codePage=x)(revision=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[0]["dnsHostName"]), "x")
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[1]["dnsHostName"]), "y")
        self.assertEqual(str(res[1]["lastLogon"]), "y")

        # Search by conjunction of remote attributes
        res = self.ldb.search(expression="(&(lastLogon=x)(description=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[1]["dnsHostName"]), "x")
        self.assertEqual(str(res[1]["lastLogon"]), "x")

        # Search by conjunction of local and remote attribute
        res = self.ldb.search(expression="(&(codePage=x)(description=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[0]["dnsHostName"]), "x")
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[1]["dnsHostName"]), "y")
        self.assertEqual(str(res[1]["lastLogon"]), "y")

        # Search by conjunction of local and remote attribute w/o match
        attrs = ["dnsHostName", "lastLogon"]
        res = self.ldb.search(expression="(&(codePage=x)(nextRid=x))",
                              attrs=attrs)
        self.assertEqual(len(res), 0)
        res = self.ldb.search(expression="(&(revision=x)(lastLogon=z))",
                              attrs=attrs)
        self.assertEqual(len(res), 0)

        # Search by disjunction of local attributes
        res = self.ldb.search(expression="(|(revision=x)(dnsHostName=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 2)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[0]["dnsHostName"]), "x")
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[1]["dnsHostName"]), "y")
        self.assertEqual(str(res[1]["lastLogon"]), "y")

        # Search by disjunction of remote attributes
        res = self.ldb.search(expression="(|(badPwdCount=x)(lastLogon=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 3)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertFalse("dnsHostName" in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertFalse("dnsHostName" in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[2]["dnsHostName"]), "x")
        self.assertEqual(str(res[2]["lastLogon"]), "x")

        # Search by disjunction of local and remote attribute
        res = self.ldb.search(expression="(|(revision=x)(lastLogon=y))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 3)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=B"))
        self.assertFalse("dnsHostName" in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "y")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[1]["dnsHostName"]), "x")
        self.assertEqual(str(res[1]["lastLogon"]), "x")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[2]["dnsHostName"]), "y")
        self.assertEqual(str(res[2]["lastLogon"]), "y")

        # Search by disjunction of local and remote attribute w/o match
        res = self.ldb.search(expression="(|(codePage=y)(nextRid=z))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 0)

        # Search by negated local attribute
        res = self.ldb.search(expression="(!(revision=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 6)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[2])
        self.assertEqual(str(res[2]["lastLogon"]), "z")
        self.assertEqual(str(res[3].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[3]["dnsHostName"]), "z")
        self.assertEqual(str(res[3]["lastLogon"]), "z")

        # Search by negated remote attribute
        res = self.ldb.search(expression="(!(description=x))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 4)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "z")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[1]["dnsHostName"]), "z")
        self.assertEqual(str(res[1]["lastLogon"]), "z")

        # Search by negated conjunction of local attributes
        res = self.ldb.search(expression="(!(&(codePage=x)(revision=x)))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 6)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[2])
        self.assertEqual(str(res[2]["lastLogon"]), "z")
        self.assertEqual(str(res[3].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[3]["dnsHostName"]), "z")
        self.assertEqual(str(res[3]["lastLogon"]), "z")

        # Search by negated conjunction of remote attributes
        res = self.ldb.search(expression="(!(&(lastLogon=x)(description=x)))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 6)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "y")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "z")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[2]["dnsHostName"]), "y")
        self.assertEqual(str(res[2]["lastLogon"]), "y")
        self.assertEqual(str(res[3].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[3]["dnsHostName"]), "z")
        self.assertEqual(str(res[3]["lastLogon"]), "z")

        # Search by negated conjunction of local and remote attribute
        res = self.ldb.search(expression="(!(&(codePage=x)(description=x)))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 6)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[2])
        self.assertEqual(str(res[2]["lastLogon"]), "z")
        self.assertEqual(str(res[3].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[3]["dnsHostName"]), "z")
        self.assertEqual(str(res[3]["lastLogon"]), "z")

        # Search by negated disjunction of local attributes
        res = self.ldb.search(expression="(!(|(revision=x)(dnsHostName=x)))",
                              attrs=["dnsHostName", "lastLogon"])
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[2])
        self.assertEqual(str(res[2]["lastLogon"]), "z")
        self.assertEqual(str(res[3].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[3]["dnsHostName"]), "z")
        self.assertEqual(str(res[3]["lastLogon"]), "z")

        # Search by negated disjunction of remote attributes
        res = self.ldb.search(expression="(!(|(badPwdCount=x)(lastLogon=x)))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 5)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "z")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=Y"))
        self.assertEqual(str(res[1]["dnsHostName"]), "y")
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[2]["dnsHostName"]), "z")
        self.assertEqual(str(res[2]["lastLogon"]), "z")

        # Search by negated disjunction of local and remote attribute
        res = self.ldb.search(expression="(!(|(revision=x)(lastLogon=y)))",
                              attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 5)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "z")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[2]["dnsHostName"]), "z")
        self.assertEqual(str(res[2]["lastLogon"]), "z")

        # Search by complex parse tree
        res = self.ldb.search(expression="(|(&(revision=x)(dnsHostName=x))(!(&(description=x)(nextRid=y)))(badPwdCount=y))", attrs=["dnsHostName", "lastLogon"])
        self.assertEqual(len(res), 7)
        res = sorted(res, key=attrgetter('dn'))
        self.assertEqual(str(res[0].dn), self.samba4.dn("cn=A"))
        self.assertTrue("dnsHostName" not in res[0])
        self.assertEqual(str(res[0]["lastLogon"]), "x")
        self.assertEqual(str(res[1].dn), self.samba4.dn("cn=B"))
        self.assertTrue("dnsHostName" not in res[1])
        self.assertEqual(str(res[1]["lastLogon"]), "y")
        self.assertEqual(str(res[2].dn), self.samba4.dn("cn=C"))
        self.assertTrue("dnsHostName" not in res[2])
        self.assertEqual(str(res[2]["lastLogon"]), "z")
        self.assertEqual(str(res[3].dn), self.samba4.dn("cn=X"))
        self.assertEqual(str(res[3]["dnsHostName"]), "x")
        self.assertEqual(str(res[3]["lastLogon"]), "x")
        self.assertEqual(str(res[4].dn), self.samba4.dn("cn=Z"))
        self.assertEqual(str(res[4]["dnsHostName"]), "z")
        self.assertEqual(str(res[4]["lastLogon"]), "z")

        # Clean up
        dns = [self.samba4.dn("cn=%s" % n) for n in ["A", "B", "C", "X", "Y", "Z"]]
        for dn in dns:
            self.ldb.delete(dn)

    def test_map_modify_local(self):
        """Modification of local records."""
        # Add local record
        dn = "cn=test,dc=idealx,dc=org"
        self.ldb.add({"dn": dn,
                      "cn": "test",
                      "foo": "bar",
                      "revision": "1",
                      "description": "test"})
        # Check it's there
        attrs = ["foo", "revision", "description"]
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["foo"]), "bar")
        self.assertEqual(str(res[0]["revision"]), "1")
        self.assertEqual(str(res[0]["description"]), "test")
        # Check it's not in the local db
        res = self.samba4.db.search(expression="(cn=test)",
                                    scope=SCOPE_DEFAULT, attrs=attrs)
        self.assertEqual(len(res), 0)
        # Check it's not in the remote db
        res = self.samba3.db.search(expression="(cn=test)",
                                    scope=SCOPE_DEFAULT, attrs=attrs)
        self.assertEqual(len(res), 0)

        # Modify local record
        ldif = """
dn: """ + dn + """
replace: foo
foo: baz
replace: description
description: foo
"""
        self.ldb.modify_ldif(ldif)
        # Check in local db
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["foo"]), "baz")
        self.assertEqual(str(res[0]["revision"]), "1")
        self.assertEqual(str(res[0]["description"]), "foo")

        # Rename local record
        dn2 = "cn=toast,dc=idealx,dc=org"
        self.ldb.rename(dn, dn2)
        # Check in local db
        res = self.ldb.search(dn2, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["foo"]), "baz")
        self.assertEqual(str(res[0]["revision"]), "1")
        self.assertEqual(str(res[0]["description"]), "foo")

        # Delete local record
        self.ldb.delete(dn2)
        # Check it's gone
        res = self.ldb.search(dn2, scope=SCOPE_BASE)
        self.assertEqual(len(res), 0)

    def test_map_modify_remote_remote(self):
        """Modification of remote data of remote records"""
        # Add remote record
        dn = self.samba4.dn("cn=test")
        dn2 = self.samba3.dn("cn=test")
        self.samba3.db.add({"dn": dn2,
                            "cn": "test",
                            "description": "foo",
                            "sambaBadPasswordCount": "3",
                            "sambaNextRid": "1001"})
        # Check it's there
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE,
                                    attrs=["description", "sambaBadPasswordCount", "sambaNextRid"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "foo")
        self.assertEqual(str(res[0]["sambaBadPasswordCount"]), "3")
        self.assertEqual(str(res[0]["sambaNextRid"]), "1001")
        # Check in mapped db
        attrs = ["description", "badPwdCount", "nextRid"]
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs, expression="")
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "foo")
        self.assertEqual(str(res[0]["badPwdCount"]), "3")
        self.assertEqual(str(res[0]["nextRid"]), "1001")
        # Check in local db
        res = self.samba4.db.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 0)

        # Modify remote data of remote record
        ldif = """
dn: """ + dn + """
replace: description
description: test
replace: badPwdCount
badPwdCount: 4
"""
        self.ldb.modify_ldif(ldif)
        # Check in mapped db
        res = self.ldb.search(dn, scope=SCOPE_BASE,
                              attrs=["description", "badPwdCount", "nextRid"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["badPwdCount"]), "4")
        self.assertEqual(str(res[0]["nextRid"]), "1001")
        # Check in remote db
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE,
                                    attrs=["description", "sambaBadPasswordCount", "sambaNextRid"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["sambaBadPasswordCount"]), "4")
        self.assertEqual(str(res[0]["sambaNextRid"]), "1001")

        # Rename remote record
        dn2 = self.samba4.dn("cn=toast")
        self.ldb.rename(dn, dn2)
        # Check in mapped db
        dn = dn2
        res = self.ldb.search(dn, scope=SCOPE_BASE,
                              attrs=["description", "badPwdCount", "nextRid"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["badPwdCount"]), "4")
        self.assertEqual(str(res[0]["nextRid"]), "1001")
        # Check in remote db
        dn2 = self.samba3.dn("cn=toast")
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE,
                                    attrs=["description", "sambaBadPasswordCount", "sambaNextRid"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["sambaBadPasswordCount"]), "4")
        self.assertEqual(str(res[0]["sambaNextRid"]), "1001")

        # Delete remote record
        self.ldb.delete(dn)
        # Check in mapped db that it's removed
        res = self.ldb.search(dn, scope=SCOPE_BASE)
        self.assertEqual(len(res), 0)
        # Check in remote db
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE)
        self.assertEqual(len(res), 0)

    def test_map_modify_remote_local(self):
        """Modification of local data of remote records"""
        # Add remote record (same as before)
        dn = self.samba4.dn("cn=test")
        dn2 = self.samba3.dn("cn=test")
        self.samba3.db.add({"dn": dn2,
                            "cn": "test",
                            "description": "foo",
                            "sambaBadPasswordCount": "3",
                            "sambaNextRid": "1001"})

        # Modify local data of remote record
        ldif = """
dn: """ + dn + """
add: revision
revision: 1
replace: description
description: test

"""
        self.ldb.modify_ldif(ldif)
        # Check in mapped db
        attrs = ["revision", "description"]
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["revision"]), "1")
        # Check in remote db
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertTrue("revision" not in res[0])
        # Check in local db
        res = self.samba4.db.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("description" not in res[0])
        self.assertEqual(str(res[0]["revision"]), "1")

        # Delete (newly) split record
        self.ldb.delete(dn)

    def test_map_modify_split(self):
        """Testing modification of split records"""
        # Add split record
        dn = self.samba4.dn("cn=test")
        dn2 = self.samba3.dn("cn=test")
        self.ldb.add({
            "dn": dn,
            "cn": "test",
            "description": "foo",
            "badPwdCount": "3",
            "nextRid": "1001",
            "revision": "1"})
        # Check it's there
        attrs = ["description", "badPwdCount", "nextRid", "revision"]
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "foo")
        self.assertEqual(str(res[0]["badPwdCount"]), "3")
        self.assertEqual(str(res[0]["nextRid"]), "1001")
        self.assertEqual(str(res[0]["revision"]), "1")
        # Check in local db
        res = self.samba4.db.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("description" not in res[0])
        self.assertTrue("badPwdCount" not in res[0])
        self.assertTrue("nextRid" not in res[0])
        self.assertEqual(str(res[0]["revision"]), "1")
        # Check in remote db
        attrs = ["description", "sambaBadPasswordCount", "sambaNextRid",
                 "revision"]
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "foo")
        self.assertEqual(str(res[0]["sambaBadPasswordCount"]), "3")
        self.assertEqual(str(res[0]["sambaNextRid"]), "1001")
        self.assertTrue("revision" not in res[0])

        # Modify of split record
        ldif = """
dn: """ + dn + """
replace: description
description: test
replace: badPwdCount
badPwdCount: 4
replace: revision
revision: 2
"""
        self.ldb.modify_ldif(ldif)
        # Check in mapped db
        attrs = ["description", "badPwdCount", "nextRid", "revision"]
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["badPwdCount"]), "4")
        self.assertEqual(str(res[0]["nextRid"]), "1001")
        self.assertEqual(str(res[0]["revision"]), "2")
        # Check in local db
        res = self.samba4.db.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("description" not in res[0])
        self.assertTrue("badPwdCount" not in res[0])
        self.assertTrue("nextRid" not in res[0])
        self.assertEqual(str(res[0]["revision"]), "2")
        # Check in remote db
        attrs = ["description", "sambaBadPasswordCount", "sambaNextRid",
                 "revision"]
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["sambaBadPasswordCount"]), "4")
        self.assertEqual(str(res[0]["sambaNextRid"]), "1001")
        self.assertTrue("revision" not in res[0])

        # Rename split record
        dn2 = self.samba4.dn("cn=toast")
        self.ldb.rename(dn, dn2)
        # Check in mapped db
        dn = dn2
        attrs = ["description", "badPwdCount", "nextRid", "revision"]
        res = self.ldb.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["badPwdCount"]), "4")
        self.assertEqual(str(res[0]["nextRid"]), "1001")
        self.assertEqual(str(res[0]["revision"]), "2")
        # Check in local db
        res = self.samba4.db.search(dn, scope=SCOPE_BASE, attrs=attrs)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn)
        self.assertTrue("description" not in res[0])
        self.assertTrue("badPwdCount" not in res[0])
        self.assertTrue("nextRid" not in res[0])
        self.assertEqual(str(res[0]["revision"]), "2")
        # Check in remote db
        dn2 = self.samba3.dn("cn=toast")
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE,
                                    attrs=["description", "sambaBadPasswordCount", "sambaNextRid",
                                           "revision"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), dn2)
        self.assertEqual(str(res[0]["description"]), "test")
        self.assertEqual(str(res[0]["sambaBadPasswordCount"]), "4")
        self.assertEqual(str(res[0]["sambaNextRid"]), "1001")
        self.assertTrue("revision" not in res[0])

        # Delete split record
        self.ldb.delete(dn)
        # Check in mapped db
        res = self.ldb.search(dn, scope=SCOPE_BASE)
        self.assertEqual(len(res), 0)
        # Check in local db
        res = self.samba4.db.search(dn, scope=SCOPE_BASE)
        self.assertEqual(len(res), 0)
        # Check in remote db
        res = self.samba3.db.search(dn2, scope=SCOPE_BASE)
        self.assertEqual(len(res), 0)



# Unix SMB/CIFS implementation.
#
# Copyright (C) Samuel Cabrero <scabrero@samba.org> 2024
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Winbind varlink service tests """

import sys
import os
import pwd
import grp

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

from samba.tests.varlink.base import VarlinkTestCase


class VarlinkGetMembershipsTests(VarlinkTestCase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def testGetMembershipsByGroup(self):
        for group in self.groups:
            full_name = "%s%s%s" % (self.domain,
                                    self.winbind_separator,
                                    group["groupname"])
            full_members_names = []
            for m in group["members"]:
                full_members_names.append("%s%s%s" % (self.domain,
                                          self.winbind_separator,
                                          m))
            vl_members = []
            with self.cli.open("io.systemd.UserDatabase", namespaced=True) as conn:
                for r in conn.GetMemberships(service=self.varlink_service,
                                             groupName=full_name,
                                             _more=True):
                    self.assertIsNotNone(r)
                    vl_members.append(r.userName)
            self.assertEqual(sorted(vl_members),
                             sorted(full_members_names))

    def testGetMembershipsByUser(self):
        for user in self.users:
            full_username = "%s%s%s" % (self.domain,
                                        self.winbind_separator,
                                        user["username"])
            pwent = pwd.getpwnam(full_username)
            glgid = os.getgrouplist(pwent.pw_name, pwent.pw_gid)
            nss_list = []
            for gid in glgid:
                grent = grp.getgrgid(gid)
                # nss_wrapper looks into files first, and "ADDOMAIN/domain users" is
                # mapped to "users" from files NSS group db.
                gname = grent.gr_name
                if gname == "users":
                    gname = "%s%s%s" % (self.domain,
                                        self.winbind_separator,
                                        "domain users")
                nss_list.append(gname)

            vl_list = []
            with self.cli.open("io.systemd.UserDatabase", namespaced=True) as conn:
                for r in conn.GetMemberships(service=self.varlink_service,
                                             userName=full_username,
                                             _more=True):
                    self.assertIsNotNone(r)
                    vl_list.append(r.groupName)

            self.assertEqual(sorted(nss_list), sorted(vl_list))

if __name__ == "__main__":
    import unittest
    unittest.main()



# Graph topology utilities, used by KCC
#
# Copyright (C) Andrew Bartlett 2015
#
# Copyright goes to Andrew Bartlett, but the actual work was performed
# by Douglas Bagnall and Garming Sam.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import itertools

from samba.graph import dot_graph


def write_dot_file(basename, edge_list, vertices=None, label=None,
                   dot_file_dir=None, debug=None, **kwargs):
    s = dot_graph(vertices, edge_list, title=label, **kwargs)
    if label:
        # sanitise DN and guid labels
        basename += '_' + label.replace(', ', '')

    filename = os.path.join(dot_file_dir, "%s.dot" % basename)
    if debug is not None:
        debug("writing graph to %s" % filename)
    f = open(filename, 'w')
    f.write(s)
    f.close()


class GraphError(Exception):
    pass


def verify_graph_complete(edges, vertices, edge_vertices):
    """The graph is complete, which is to say there is an edge between
    every pair of nodes."""
    for v in vertices:
        remotes = set()
        for a, b in edges:
            if a == v:
                remotes.add(b)
            elif b == v:
                remotes.add(a)
        if len(remotes) + 1 != len(vertices):
            raise GraphError("graph is not fully connected")


def verify_graph_connected(edges, vertices, edge_vertices):
    """There is a path between any two nodes."""
    if not edges:
        if len(vertices) <= 1:
            return
        raise GraphError("all vertices are disconnected because "
                         "there are no edges:")

    remaining_edges = list(edges)
    reached = set(remaining_edges.pop())
    while True:
        doomed = []
        for i, e in enumerate(remaining_edges):
            a, b = e
            if a in reached:
                reached.add(b)
                doomed.append(i)
            elif b in reached:
                reached.add(a)
                doomed.append(i)
        if not doomed:
            break
        for i in reversed(doomed):
            del remaining_edges[i]

    if remaining_edges or reached != set(vertices):
        s = ("the graph is not connected, "
             "as the following vertices are unreachable:\n ")
        s += '\n '.join(v for v in sorted(vertices)
                        if v not in reached)
        raise GraphError(s)


def verify_graph_connected_under_edge_failures(edges, vertices, edge_vertices):
    """The graph stays connected when any single edge is removed."""
    if len(edges) == 0:
        return verify_graph_connected(edges, vertices, edge_vertices)

    for subset in itertools.combinations(edges, len(edges) - 1):
        try:
            verify_graph_connected(subset, vertices, edge_vertices)
        except GraphError as e:
            for edge in edges:
                if edge not in subset:
                    raise GraphError("The graph will be disconnected when the "
                                     "connection from %s to %s fails" % edge)


def verify_graph_connected_under_vertex_failures(edges, vertices,
                                                 edge_vertices):
    """The graph stays connected when any single vertex is removed."""
    for v in vertices:
        sub_vertices = [x for x in vertices if x is not v]
        sub_edges = [x for x in edges if v not in x]
        verify_graph_connected(sub_edges, sub_vertices, sub_vertices)


def verify_graph_forest(edges, vertices, edge_vertices):
    """The graph contains no loops."""
    trees = [set(e) for e in edges]
    while True:
        for a, b in itertools.combinations(trees, 2):
            intersection = a & b
            if intersection:
                if len(intersection) == 1:
                    a |= b
                    trees.remove(b)
                    break
                else:
                    raise GraphError("there is a loop in the graph\n"
                                     " vertices %s\n edges %s\n"
                                     " intersection %s" %
                                     (vertices, edges, intersection))
        else:
            # no break in itertools.combinations loop means no
            # further mergers, so we're done.
            #
            # XXX here we also know whether it is a tree or a
            # forest by len(trees) but the connected test already
            # tells us that.
            return


def verify_graph_multi_edge_forest(edges, vertices, edge_vertices):
    """This allows a forest with duplicate edges. That is if multiple
    edges go between the same two vertices, they are treated as a
    single edge by this test.

    e.g.:
                        o
    pass: o-o=o  o=o   (|)             fail:  o-o
            `o          o                     `o'
    """
    unique_edges = set(edges)
    trees = [set(e) for e in unique_edges]
    while True:
        for a, b in itertools.combinations(trees, 2):
            intersection = a & b
            if intersection:
                if len(intersection) == 1:
                    a |= b
                    trees.remove(b)
                    break
                else:
                    raise GraphError("there is a loop in the graph")
        else:
            return


def verify_graph_no_lonely_vertices(edges, vertices, edge_vertices):
    """There are no vertices without edges."""
    lonely = set(vertices) - set(edge_vertices)
    if lonely:
        raise GraphError("some vertices are not connected:\n%s" %
                         '\n'.join(sorted(lonely)))


def verify_graph_no_unknown_vertices(edges, vertices, edge_vertices):
    """The edge endpoints contain no vertices that are otherwise unknown."""
    unknown = set(edge_vertices) - set(vertices)
    if unknown:
        raise GraphError("some edge vertices are seemingly unknown:\n%s" %
                         '\n'.join(sorted(unknown)))


def verify_graph_directed_double_ring(edges, vertices, edge_vertices):
    """Each node has at least two directed edges leaving it, and two
    arriving. The edges work in pairs that have the same end points
    but point in opposite directions. The pairs form a path that
    touches every vertex and form a loop.

    There might be other connections that *aren't* part of the ring.

    Deciding this for sure is NP-complete (the Hamiltonian path
    problem), but there are some easy failures that can be detected.
    So far we check for:
      - leaf nodes
      - disjoint subgraphs
      - robustness against edge and vertex failure
    """
    # a zero or one node graph is OK with no edges.
    # The two vertex case is special. Use
    # verify_graph_directed_double_ring_or_small() to allow that.
    if not edges and len(vertices) <= 1:
        return
    if len(edges) < 2 * len(vertices):
        raise GraphError("directed double ring requires at least twice "
                         "as many edges as vertices")

    # Reduce the problem space by looking only at bi-directional links.
    half_duplex = set(edges)
    duplex_links = set()
    for edge in edges:
        rev_edge = (edge[1], edge[0])
        if edge in half_duplex and rev_edge in half_duplex:
            duplex_links.add(edge)
            half_duplex.remove(edge)
            half_duplex.remove(rev_edge)

    # the Hamiltonian cycle problem is NP-complete in general, but we
    # can cheat a bit and prove a less strong result.
    #
    # We declutter the graph by replacing nodes with edges connecting
    # their neighbours.
    #
    #       A-B-C --> A-C
    #
    #    -A-B-C-   -->  -A--C-
    #       `D_           `D'_
    #
    # In the end there should be a single 2 vertex graph.

    edge_map = {}
    for a, b in duplex_links:
        edge_map.setdefault(a, set()).add(b)
        edge_map.setdefault(b, set()).add(a)

    # an easy to detect failure is a lonely leaf node
    for vertex, neighbours in edge_map.items():
        if len(neighbours) == 1:
            raise GraphError("wanted double directed ring, found a leaf node"
                             "(%s)" % vertex)

    for vertex in list(edge_map.keys()):
        nset = edge_map[vertex]
        if not nset:
            continue
        for n in nset:
            n_neighbours = edge_map[n]
            n_neighbours.remove(vertex)
            n_neighbours.update(x for x in nset if x != n)
        del edge_map[vertex]

    if len(edge_map) > 1:
        raise GraphError("wanted double directed ring, but "
                         "this looks like a split graph\n"
                         "(%s can't reach each other)" %
                         ', '.join(edge_map.keys()))

    verify_graph_connected_under_edge_failures(duplex_links, vertices,
                                               edge_vertices)
    verify_graph_connected_under_vertex_failures(duplex_links, vertices,
                                                 edge_vertices)


def verify_graph_directed_double_ring_or_small(edges, vertices, edge_vertices):
    """This performs the directed_double_ring test but makes special
    concessions for small rings where the strict rules don't really
    apply."""
    if len(vertices) < 2:
        return
    if len(vertices) == 2:
        """With 2 nodes there should be a single link in each directions."""
        if (len(edges) == 2 and
            edges[0][0] == edges[1][1] and
            edges[0][1] == edges[1][0]):
            return
        raise GraphError("A two vertex graph should have an edge each way.")

    return verify_graph_directed_double_ring(edges, vertices, edge_vertices)


def verify_graph(edges, vertices=None, directed=False, properties=()):
    errors = []
    properties = [x.replace(' ', '_') for x in properties]

    edge_vertices = set()
    for a, b in edges:
        edge_vertices.add(a)
        edge_vertices.add(b)

    if vertices is None:
        vertices = edge_vertices
    else:
        vertices = set(vertices)

    for p in properties:
        fn = 'verify_graph_%s' % p
        f = globals()[fn]
        try:
            f(edges, vertices, edge_vertices)
        except GraphError as e:
            errors.append((p, e, f.__doc__))

    return errors


def verify_and_dot(basename, edges, vertices=None, label=None,
                   reformat_labels=True, directed=False,
                   properties=(), fatal=True, debug=None,
                   verify=True, dot_file_dir=None,
                   edge_colors=None, edge_labels=None,
                   vertex_colors=None):

    if dot_file_dir is not None:
        write_dot_file(basename, edges, vertices=vertices, label=label,
                       dot_file_dir=dot_file_dir,
                       reformat_labels=reformat_labels, directed=directed,
                       debug=debug, edge_colors=edge_colors,
                       edge_labels=edge_labels, vertex_colors=vertex_colors)

    if verify:
        errors = verify_graph(edges, vertices,
                              properties=properties)
        if errors:
            title = '%s %s' % (basename, label or '')
            debug("%s FAILED:" % title)
            for p, e, doc in errors:
                debug(" %18s: %s" % (p, e))
            if fatal:
                raise GraphError("The '%s' graph lacks the following "
                                 "properties:\n%s" %
                                 (title, '\n'.join('%s: %s' % (p, e)
                                                   for p, e, doc in errors)))


def list_verify_tests():
    for k, v in sorted(globals().items()):
        if k.startswith('verify_graph_'):
            print(k.replace('verify_graph_', ''))
            if v.__doc__:
                print('    %s' % (v.__doc__.rstrip()))
            else:
                print()



# Tests for SamDb password change audit logging.
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for the SamDb logging of password changes.
"""

import samba.tests
from samba.dcerpc.messaging import MSG_DSDB_PWD_LOG, DSDB_PWD_EVENT_NAME
from samba.samdb import SamDB
from samba.auth import system_session
import os
from samba.tests.audit_log_base import AuditLogTestBase
from samba.tests import delete_force
from samba.net import Net
from ldb import (LdbError,
                 ERR_INSUFFICIENT_ACCESS_RIGHTS,
                 ERR_INVALID_DN_SYNTAX,
                 ERR_NO_SUCH_ATTRIBUTE)
from samba.dcerpc.windows_event_ids import (
    EVT_ID_PASSWORD_CHANGE,
    EVT_ID_PASSWORD_RESET,
    EVT_ID_DIRECTORY_OBJECT_CHANGE
)


USER_NAME = "auditlogtestuser"
USER_PASS = samba.generate_random_password(32, 32)

SECOND_USER_NAME = "auditlogtestuser02"
SECOND_USER_PASS = samba.generate_random_password(32, 32)

MACHINE_NAME = "auditlogtestmachineuser"
MACHINE_PASS = samba.generate_random_password(32, 32)


class AuditLogPassChangeTests(AuditLogTestBase):

    def setUp(self):
        self.message_type = MSG_DSDB_PWD_LOG
        self.event_type = DSDB_PWD_EVENT_NAME
        super().setUp()

        self.server_ip = os.environ["SERVER_IP"]

        host = "ldap://%s" % os.environ["SERVER"]
        self.ldb = SamDB(url=host,
                         session_info=system_session(),
                         credentials=self.get_credentials(),
                         lp=self.get_loadparm())
        self.server = os.environ["SERVER"]

        # Gets back the basedn
        self.base_dn = self.ldb.domain_dn()

        # Get the old "dSHeuristics" if it was set
        dsheuristics = self.ldb.get_dsheuristics()

        # Set the "dSHeuristics" to activate the correct "userPassword"
        # behaviour
        self.ldb.set_dsheuristics("000000001")

        # Reset the "dSHeuristics" as they were before
        self.addCleanup(self.ldb.set_dsheuristics, dsheuristics)

        # Get the old "minPwdAge"
        minPwdAge = self.ldb.get_minPwdAge()

        # Set it temporarily to "0"
        self.ldb.set_minPwdAge("0")
        self.base_dn = self.ldb.domain_dn()

        # Reset the "minPwdAge" as it was before
        self.addCleanup(self.ldb.set_minPwdAge, minPwdAge)

        # (Re)adds the test user USER_NAME with password USER_PASS
        delete_force(self.ldb, "cn=" + USER_NAME + ",cn=users," + self.base_dn)
        delete_force(
            self.ldb,
            "cn=" + SECOND_USER_NAME + ",cn=users," + self.base_dn)
        self.ldb.add({
            "dn": "cn=" + USER_NAME + ",cn=users," + self.base_dn,
            "objectclass": "user",
            "sAMAccountName": USER_NAME,
            "userPassword": USER_PASS
        })

        # (Re)adds the test user MACHINE_NAME with password MACHINE_PASS
        delete_force(
            self.ldb,
            "cn=" + MACHINE_NAME + ",cn=users," + self.base_dn)
        self.ldb.add({
            "dn": "cn=" + MACHINE_NAME + ",cn=users," + self.base_dn,
            "objectclass": "computer",
            "sAMAccountName": MACHINE_NAME,
            "userPassword": MACHINE_PASS
        })

    #
    # Discard the messages from the setup code
    #
    def discardSetupMessages(self, dn):
        self.waitForMessages(1, dn=dn)
        self.discardMessages()

    def test_net_change_password(self):

        dn = "CN=" + USER_NAME + ",CN=Users," + self.base_dn
        self.discardSetupMessages(dn)

        creds = self.insta_creds(template=self.get_credentials())

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server)
        password = "newPassword!!42"

        net.change_password(newpassword=password,
                            username=USER_NAME,
                            oldpassword=USER_PASS)

        messages = self.waitForMessages(1, net, dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")
        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_CHANGE, audit["eventId"])
        self.assertEqual("Change", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "DCE/RPC")
        self.assertTrue(self.is_guid(audit["transactionId"]))

    def test_net_set_password_user_without_permission(self):

        dn = "CN=" + USER_NAME + ",CN=Users," + self.base_dn
        self.discardSetupMessages(dn)

        self.ldb.newuser(SECOND_USER_NAME, SECOND_USER_PASS)

        #
        # Get the password reset from the user add
        #
        dn = "CN=" + SECOND_USER_NAME + ",CN=Users," + self.base_dn
        messages = self.waitForMessages(1, dn=dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")

        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_RESET, audit["eventId"])
        self.assertEqual("Reset", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "LDAP")
        self.assertTrue(self.is_guid(audit["transactionId"]))
        self.assertEqual(0, audit["statusCode"])
        self.assertEqual("Success", audit["status"])
        self.discardMessages()

        creds = self.insta_creds(
            template=self.get_credentials(),
            username=SECOND_USER_NAME,
            userpass=SECOND_USER_PASS,
            kerberos_state=None)

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server)
        password = "newPassword!!42"
        domain = lp.get("workgroup")

        try:
            net.set_password(newpassword=password,
                             account_name=USER_NAME,
                             domain_name=domain)
            self.fail("Expected exception not thrown")
        except Exception:
            pass

        dn = "CN=" + USER_NAME + ",CN=Users," + self.base_dn
        messages = self.waitForMessages(1, net, dn=dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")

        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_RESET, audit["eventId"])
        self.assertEqual("Reset", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "DCE/RPC")
        self.assertTrue(self.is_guid(audit["transactionId"]))
        self.assertEqual(ERR_INSUFFICIENT_ACCESS_RIGHTS, audit["statusCode"])
        self.assertEqual("insufficient access rights", audit["status"])

    def test_net_set_password(self):

        dn = "CN=" + USER_NAME + ",CN=Users," + self.base_dn
        self.discardSetupMessages(dn)

        creds = self.insta_creds(template=self.get_credentials())

        lp = self.get_loadparm()
        net = Net(creds, lp, server=self.server)
        password = "newPassword!!42"
        domain = lp.get("workgroup")

        net.set_password(newpassword=password,
                         account_name=USER_NAME,
                         domain_name=domain)

        dn = "CN=" + USER_NAME + ",CN=Users," + self.base_dn
        messages = self.waitForMessages(1, net, dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")

        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_RESET, audit["eventId"])
        self.assertEqual("Reset", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "DCE/RPC")
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        self.assertTrue(self.is_guid(audit["transactionId"]))

    def test_ldap_change_password(self):

        dn = "cn=" + USER_NAME + ",cn=users," + self.base_dn
        self.discardSetupMessages(dn)

        new_password = samba.generate_random_password(32, 32)
        self.ldb.modify_ldif(
            "dn: " + dn + "\n" +
            "changetype: modify\n" +
            "delete: userPassword\n" +
            "userPassword: " + USER_PASS + "\n" +
            "add: userPassword\n" +
            "userPassword: " + new_password + "\n")

        messages = self.waitForMessages(1, dn=dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")

        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_CHANGE, audit["eventId"])
        self.assertEqual("Change", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        self.assertTrue(self.is_guid(audit["sessionId"]))
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "LDAP")
        self.assertTrue(self.is_guid(audit["transactionId"]))

    def test_ldap_replace_password(self):

        dn = "cn=" + USER_NAME + ",cn=users," + self.base_dn
        self.discardSetupMessages(dn)

        new_password = samba.generate_random_password(32, 32)
        self.ldb.modify_ldif(
            "dn: " + dn + "\n" +
            "changetype: modify\n" +
            "replace: userPassword\n" +
            "userPassword: " + new_password + "\n")

        messages = self.waitForMessages(1, dn=dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")

        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_RESET, audit["eventId"])
        self.assertEqual("Reset", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        self.assertTrue(self.is_guid(audit["sessionId"]))
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "LDAP")
        self.assertTrue(self.is_guid(audit["transactionId"]))

    def test_ldap_add_user(self):

        # The setup code adds a user, so we check for the password event
        # generated by it.
        dn = "cn=" + USER_NAME + ",cn=users," + self.base_dn
        messages = self.waitForMessages(1, dn=dn)
        print("Received %d messages" % len(messages))
        self.assertEqual(1,
                          len(messages),
                          "Did not receive the expected number of messages")

        #
        # The first message should be the reset from the Setup code.
        #
        audit = messages[0]["passwordChange"]
        self.assertEqual(EVT_ID_PASSWORD_RESET, audit["eventId"])
        self.assertEqual("Reset", audit["action"])
        self.assertEqual(dn, audit["dn"])
        self.assertRegex(audit["remoteAddress"],
                         self.remoteAddress)
        session_id = self.get_session()
        self.assertEqual(session_id, audit["sessionId"])
        service_description = self.get_service_description()
        self.assertEqual(service_description, "LDAP")
        self.assertTrue(self.is_guid(audit["sessionId"]))
        self.assertTrue(self.is_guid(audit["transactionId"]))


    # These are some valid BinaryDN prefixes for
    # msDS-KeyCredentialLink values.
    kcl_prefixes = [
        ("B:772:000200002000012849ABF23C432F5428756A641827A6CD2E4A4F40"
         "9F4DE57BA152B7C2F31E46EC200002A5499B70EC0FBEF4BCB78AF3878D08"
         "2B2D43527BBAA230462707947C128BAF4226010330820122300D06092A86"
         "4886F70D01010105000382010F003082010A02820101009A16283EA51D94"
         "11FD54C73615D90E4B3B8DAF152D699A79C3B217DDBA7787DE419E3E31E6"
         "012088FD9B4EC0EA9199E91EBE2F99F48A4E7084D819CF76C67C58F5050E"
         "E5428B09676CA475EADFAD91FBEF2835820BE4CDE3F3A3B33DB6A0C75956"
         "47CA8489566EBCF6B748DC637EC2D34D255E008DAF93F3DC77478B7526DE"
         "5A6B4638FC50A622233C2BBA935A0E2BBA6DDCE32D539E57AD12FAA7BA33"
         "1C92D32E54E05290132E66D25476458FDDFBF4940DA97F19B63C30EFE0EF"
         "1F77A2751CD8EAA7EF6DFA54FF0B3500D9E66A1ECC9D435FF0206EA27047"
         "8A37CD9F949EB6CA82925A138C5852C5C763AD6EA60003FCE5C7D486382B"
         "8E347B8CF6645F449902030100010100040108000900CB9A52E817DC01:"),

        ("B:772:00020000200001E251886DDA8335FA6A9B311F1FE4ABCDA7C04F05"
         "33668BA7B143760C675703BB200002E324A0CCE2393C603686A0C7828B53"
         "77A9B5E8C639131F657C3CEB39EDED7D4F26010330820122300D06092A86"
         "4886F70D01010105000382010F003082010A0282010100A5DF595B4E0F36"
         "9A4EF8CD5A2F6AF9322460C14175DDB7DE5F3A494A7591EC1096A5150800"
         "F80401F7D8739C8900165035E231E85AA0E39A9C75BE760764BCDC82A4CB"
         "955D84AFA47A8CACC35BAB6775478BF214A81726263D79CAA1623EA60DAB"
         "2C61CAE29905FE4A89467736F47ADCA27170DBC77F5A82E28075E4D44FB9"
         "76FFF5400B1AFC21B57C52718864253D57E780DEF8F40C990D30FB253831"
         "BC95B01EEFFB312E52F7B773CFB1FF9E48A40B9C82E6B176464087C862F5"
         "EA1D38809CF6454A8F169B993FFA857D1D928E4488EB13C947EE847D30F1"
         "7CF1FAA85DE6AFD07ED82504C024E0CEA5B47E02515E57C3258963E80137"
         "15EB0D3B126067A79F02030100010100040108000900CB9A52E817DC01:"),

        ("B:772:0002000020000118ADEEC2CBCC63FCA44F332CAE39275E12CB881F"
         "68B582ED2F3982718832126120000269F60B033A8697DB69835A356C9263"
         "1CF87FDE6E448F3426285B3978E85C385C26010330820122300D06092A86"
         "4886F70D01010105000382010F003082010A0282010100CADFE83AC8FCE1"
         "B7A999AE162DC6BD6CC53D686E6F0CB866AFADCF64F736249E2CB4F438E2"
         "78B636C4151C540BC4677821E66E6D88E6875A6B9B4F473D41C1C05376A4"
         "F929076E515698F076F1F1BBEB25AAE062C9973D6436E3E4B48F74B5C0A5"
         "B1FAD5026B8AB9E021849FFF8D18E54643C5F5FB57590BDC1CBC747C53CE"
         "A696D5379EE3678E8AE015E2AE4AEA5B79F03E953D415986197F00D876E1"
         "D23A28CC944901FCE5A902C9671CC46D7AA7F349F4F80CEF74FF035AEAEE"
         "30D6FD8F907B08C9A23031B9E44AC84A4010BFA06A468D94B034DDF202A7"
         "1B20663864703AC8205CE7C60B5FC18253C94CB8389381E05DB70B08BCCC"
         "EE58A5E81BA2FE18AD02030100010100040108000980613353E817DC01:"),
    ]

    def test_ldap_key_credential_link(self):
        """Test logging of msDS-KeyCredentialLink public key changes.
        """
        # The restrictions on msDS-KeyCredentialLink changes work
        # different from those on the password attributes. We don't
        # care about dsHeuristics, minPwdAge, etc, but we do care
        # about things like GUID_DRS_DS_VALIDATED_WRITE_COMPUTER and
        # ACLs.
        #
        # To avoid all the set-up cost of making a fresh DB and user,
        # we use sub-tests in this test.
        #
        # See source4/dsdb/tests/pythpn/key_credential_link.py for
        # deeper tests of the restrictions on setting this attribute.

        dn = f"cn={USER_NAME},cn=users,{self.base_dn}"
        self.discardSetupMessages(dn)

        kcls = [f"{x}{dn}" for x in self.kcl_prefixes]
        session_id = self.get_session()
        service_description = self.get_service_description()
        self.assertEqual(service_description, "LDAP")

        self._test_ldap_authentication_information(
            "msDS-keyCredentialLink", kcls)

        transactions_seen = set()

        with self.subTest("add bad KCL DN value"):
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                "replace: msDS-keyCredentialLink\n"
                f"msDS-keyCredentialLink: B:4:f1ea:{dn}\n")
            messages = self.waitForMessages(1, dn=dn)
            self.discardMessages()
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            self.assertEqual(0, audit["statusCode"])
            transactions_seen.add(audit["transactionId"])

        with self.subTest("add a second DN value"):
            # should this fail?
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                "add: msDS-keyCredentialLink\n"
                f"msDS-keyCredentialLink: B:4:f1ee:{dn}\n")
            messages = self.waitForMessages(1, dn=dn)
            self.discardMessages()
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            self.assertEqual(0, audit["statusCode"])
            transactions_seen.add(audit["transactionId"])

        # these should all have been separate transactions
        with self.subTest("check transactions"):
            self.assertEqual(len(transactions_seen), 2)
            for t in transactions_seen:
                self.assertTrue(self.is_guid(t))

        with self.subTest("add bad Binary DN value"):
            for bad_dn in ('B:6:f1ea:{dn}', 'flea', dn):
                with self.assertRaises(LdbError) as e:
                    self.ldb.modify_ldif(
                        f"dn: {dn}\n"
                        "changetype: modify\n"
                        "replace: msDS-keyCredentialLink\n"
                        f"msDS-keyCredentialLink: {bad_dn}\n")
                self.assertEqual(e.exception.args[0], ERR_INVALID_DN_SYNTAX)
            # no messages from those the 3 bad DNs
            # because DN syntax check comes first
            messages = self.waitForMessages(1, dn=dn)
            self.assertEqual(0, len(messages))


    def test_ldap_altSecurityIdentities(self):
        """Test logging of altSecurityIdentities changes.
        """
        values = [
            "X509:<SKI>123456789123",
            "X509:<S>SubjectName<I>IssuerName",
            "X509:<I>IssuerName<SR>123456789123"
        ]
        self._test_ldap_authentication_information(
            "altSecurityIdentities", values)


    def test_ldap_service_principal_name(self):
        """Test logging of servicePrincipalName changes.
        """
        values = [
            "HOST/principal1",
            "HOST/principal2",
            "HOST/Principla3"
        ]
        self._test_ldap_authentication_information(
            "servicePrincipalName", values)


    def test_ldap_dns_host_name(self):
        """Test logging of dNSHostName changes.
        """
        values = [
            "host1.test.samba.org",
            "host2.test.samba.org",
            "host3.test.samba.org"
        ]
        self._test_ldap_authentication_information(
            "dNSHostName", values, user=MACHINE_NAME)

    def test_ldap_msDS_AdditionalDnsHostName(self):
        """Test logging of msDS-AdditionalDnsHostName changes.
        """
        values = [
            "host1.test.samba.org",
            "host2.test.samba.org",
            "host3.test.samba.org"
        ]
        self._test_ldap_authentication_information(
            "msDS-AdditionalDnsHostName", values, user=MACHINE_NAME)

    def _test_ldap_authentication_information(
            self,
            attribute,
            values,
            user=USER_NAME ):
        """Test logging of authentication information changes.
        """
        #
        # To avoid all the set-up cost of making a fresh DB and user,
        # we use sub-tests in this test.
        #

        dn = f"cn={user},cn=users,{self.base_dn}"
        self.discardSetupMessages(dn)

        session_id = self.get_session()
        service_description = self.get_service_description()
        self.assertEqual(service_description, "LDAP")
        transactions_seen = set()

        with self.subTest("initial setup"):
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                f"add: {attribute}\n"
                f"{attribute}: {values[0]}\n")
            messages = self.waitForMessages(1, dn=dn)
            print("Received %d messages" % len(messages))
            self.assertEqual(1, len(messages))
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            transactions_seen.add(audit["transactionId"])
            self.assertEqual(0, audit["statusCode"])
            self.discardMessages()

        with self.subTest("replace"):
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                f"replace: {attribute}\n"
                f"{attribute}: {values[1]}\n")
            messages = self.waitForMessages(1, dn=dn)
            print("Received %d messages" % len(messages))
            self.assertEqual(1, len(messages))
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertTrue(self.is_guid(audit["sessionId"]))
            transactions_seen.add(audit["transactionId"])
            self.discardMessages()

        with self.subTest("constrained replace"):
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                f"delete: {attribute}\n"
                f"{attribute}: {values[1]}\n"
                f"add: {attribute}\n"
                f"{attribute}: {values[2]}\n")
            messages = self.waitForMessages(1, dn=dn)
            print("Received %d messages" % len(messages))
            self.assertEqual(1, len(messages))
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            self.assertEqual(0, audit["statusCode"])
            transactions_seen.add(audit["transactionId"])
            self.discardMessages()

        with self.subTest("identical replace"):
            # replacing the value with itself still sends the message.
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                f"replace: {attribute}\n"
                f"{attribute}: {values[2]}\n")
            messages = self.waitForMessages(1, dn=dn)
            print("Received %d messages" % len(messages))
            self.assertEqual(1, len(messages))
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            self.assertEqual(0, audit["statusCode"])
            transactions_seen.add(audit["transactionId"])
            self.discardMessages()

        with self.subTest("replace authentication information AND password"):
            # there should be two messages
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                f"replace:{attribute}\n"
                f"{attribute}: {values[0]}\n"
                "replace: userPassword\n"
                "userPassword: gruffalo3.\n")
            messages = self.waitForMessages(2, dn=dn)
            self.assertEqual(2, len(messages))
            pwd_audit = messages[0]["passwordChange"]
            kcl_audit = messages[1]["passwordChange"]
            # we send the password message first, but we don't need to
            # depend on that.
            if pwd_audit["eventId"] == EVT_ID_DIRECTORY_OBJECT_CHANGE:
                kcl_audit, pwd_audit = pwd_audit, kcl_audit
            del audit
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, kcl_audit["eventId"])
            self.assertEqual(EVT_ID_PASSWORD_RESET, pwd_audit["eventId"])
            self.assertEqual("Auth info change", kcl_audit["action"])
            self.assertEqual("Reset", pwd_audit["action"])
            # if we delete the action and eventId, the rest of
            # structures should be the same (sessionId, transactionId,
            # version, etc). Timestamps are in the outer message.
            del pwd_audit["eventId"]
            del pwd_audit["action"]
            del kcl_audit["eventId"]
            del kcl_audit["action"]
            # replacing an authentication information value with itself
            # still sends the message.
            self.assertEqual(kcl_audit, pwd_audit)
            transactions_seen.add(pwd_audit["transactionId"])
            self.discardMessages()

        with self.subTest("delete"):
            self.ldb.modify_ldif(
                f"dn: {dn}\n"
                "changetype: modify\n"
                f"delete: {attribute}\n"
                f"{attribute}: {values[0]}\n")
            messages = self.waitForMessages(1, dn=dn)
            print("Received %d messages" % len(messages))
            self.assertEqual(1, len(messages))
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            self.assertEqual(0, audit["statusCode"])
            transactions_seen.add(audit["transactionId"])
            self.discardMessages()

        with self.subTest("delete that which does not exist"):
            # still sends a message
            with self.assertRaises(LdbError) as e:
                self.ldb.modify_ldif(
                    f"dn: {dn}\n"
                    "changetype: modify\n"
                    f"delete: {attribute}\n"
                    f"{attribute}: {values[2]}\n")
            self.assertEqual(e.exception.args[0], ERR_NO_SUCH_ATTRIBUTE)
            messages = self.waitForMessages(1, dn=dn)
            print("Received %d messages" % len(messages))
            # We still get the message on a failed attempt
            self.assertEqual(1, len(messages))
            audit = messages[0]["passwordChange"]
            self.assertEqual(EVT_ID_DIRECTORY_OBJECT_CHANGE, audit["eventId"])
            self.assertEqual("Auth info change", audit["action"])
            self.assertEqual(dn, audit["dn"])
            self.assertIn(self.remoteAddress, audit["remoteAddress"])
            self.assertEqual(session_id, audit["sessionId"])
            transactions_seen.add(audit["transactionId"])
            self.discardMessages()
            with self.subTest("check status code"):
                self.assertEqual(ERR_NO_SUCH_ATTRIBUTE, audit["statusCode"])
                self.assertEqual("No such attribute", audit["status"])

        # these should all have been separate transactions
        with self.subTest("check transactions"):
            self.assertEqual(len(transactions_seen), 7)
            for t in transactions_seen:
                self.assertTrue(self.is_guid(t))




# Unix SMB/CIFS implementation.
#
# Blackbox tests for GMSA workflow.
#
# Copyright (C) Catalyst.Net Ltd. 2024
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import os
import sys
from shlex import quote

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

from samba.domain.models import Computer
from samba.tests import BlackboxProcessError, BlackboxTestCase, connect_samdb

DC_SERVER = os.environ["SERVER"]
SERVER = os.environ["SERVER"]
SERVER_USERNAME = os.environ["USERNAME"]
SERVER_PASSWORD = os.environ["PASSWORD"]

HOST = f"ldap://{SERVER}"
ADMIN_CREDS = f"-U{SERVER_USERNAME}%{SERVER_PASSWORD}"


class GMSABlackboxTest(BlackboxTestCase):
    """Blackbox tests for GMSA management."""

    @classmethod
    def setUpClass(cls):
        cls.lp = cls.get_loadparm()
        cls.env_creds = cls.get_env_credentials(lp=cls.lp,
                                                env_username="USERNAME",
                                                env_password="PASSWORD",
                                                env_domain="DOMAIN",
                                                env_realm="REALM")
        cls.samdb = connect_samdb(HOST, lp=cls.lp, credentials=cls.env_creds)
        super().setUpClass()

    def getpassword(self, account_name, attrs, creds=ADMIN_CREDS):
        cmd = f"samba-tool user getpassword --attributes={quote(attrs)} {account_name} -H {HOST} {creds}"
        ldif = self.check_output(cmd).decode()
        res = self.samdb.parse_ldif(ldif)
        _, user_message = next(res)

        # check each attr is returned
        for attr in attrs.split(","):
            if attr not in user_message:
                raise KeyError

        return user_message

    def test_gmsa_password_access(self):
        """Test machine account read password access."""
        machine_account = "Machine_Account$"
        machine_password = "T3stPassword0nly"
        machine_creds = f"-U{machine_account}%{machine_password}"
        gmsa_account = "GMSA_Test_User$"

        # Create a machine account and set the password.
        self.check_run(f"samba-tool computer create {machine_account} -H {HOST} {ADMIN_CREDS}")
        self.addCleanup(self.run_command, f"samba-tool computer delete {machine_account} -H {HOST} {ADMIN_CREDS}")
        self.check_run(f"samba-tool user setpassword {machine_account} --newpassword={machine_password} -H {HOST} {ADMIN_CREDS}")

        # Create a Group Managed Service Account with default SDDL.
        self.check_run(f"samba-tool service-account create --name={gmsa_account} --dns-host-name=example.com --managed-password-interval=1 -H {HOST} {ADMIN_CREDS}")
        self.addCleanup(self.run_command, f"samba-tool service-account delete --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")

        # Grant password read access to the machine account.
        self.check_run(f"samba-tool service-account group-msa-membership add --name={gmsa_account} --principal={machine_account} -H {HOST} {ADMIN_CREDS}")

        try:
            self.getpassword(gmsa_account, "unicodePwd", creds=machine_creds)
        except KeyError:
            self.fail("Failed to get unicodePwd despite being in the gMSA membership")

        # Remove password read access from the machine account and verify.
        self.check_run(f"samba-tool service-account group-msa-membership remove --name={gmsa_account} --principal={machine_account} -H {HOST} {ADMIN_CREDS}")

        try:
            self.assertRaises(KeyError, self.getpassword, gmsa_account, "unicodePwd", creds=machine_creds)
        except BlackboxProcessError:
            self.fail("Unexpected subcommand failure retrieving unicodePwd")

    def test_gmsa_add_sid_only_viewer(self):
        """Add unknown SID to password viewers and check group-msa-membership show output."""
        gmsa_account = "GMSA_Test_User$"
        unknown_sid = f"{self.samdb.domain_sid}-9999"

        self.check_run(f"samba-tool service-account create --name={gmsa_account} --dns-host-name=example.com --managed-password-interval=1 -H {HOST} {ADMIN_CREDS}")
        self.addCleanup(self.run_command, f"samba-tool service-account delete --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")

        self.check_run(f"samba-tool service-account group-msa-membership add --name={gmsa_account} --principal={unknown_sid} -H {HOST} {ADMIN_CREDS}")

        out = self.check_output(f"samba-tool service-account group-msa-membership show --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")
        self.assertIn(f"<SID={unknown_sid}>", out.decode())

    def test_custom_sddl_as_list(self):
        """Test custom SDDL that can be represented by a simple list."""
        machine_account = "Machine_Account$"
        machine_password = "T3stPassword0nly"
        gmsa_account = "GMSA_Test_User$"
        unknown_sid = f"{self.samdb.domain_sid}-9999"

        # Create a machine account and set the password.
        self.check_run(f"samba-tool computer create {machine_account} -H {HOST} {ADMIN_CREDS}")
        self.addCleanup(self.run_command, f"samba-tool computer delete {machine_account} -H {HOST} {ADMIN_CREDS}")
        self.check_run(f"samba-tool user setpassword {machine_account} --newpassword={machine_password} -H {HOST} {ADMIN_CREDS}")

        # Create GMSA with custom SDDL this time rather than the command default.
        initial_sddl = f"O:SYD:(A;;RP;;;{self.samdb.connecting_user_sid})"
        self.check_run(f'samba-tool service-account create --name={gmsa_account} --dns-host-name=example.com --group-msa-membership="{initial_sddl}" --managed-password-interval=1 -H {HOST} {ADMIN_CREDS}')
        self.addCleanup(self.run_command, f"samba-tool service-account delete --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")

        # Read the SDDL using service-account view JSON, it should be the same.
        out = self.check_output(f"samba-tool service-account view --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")
        gmsa = json.loads(out.decode())
        self.assertEqual(gmsa["msDS-GroupMSAMembership"], initial_sddl)

        # Add the machine account as a password viewer.
        self.check_run(f"samba-tool service-account group-msa-membership add --name={gmsa_account} --principal={machine_account} -H {HOST} {ADMIN_CREDS}")

        # Add the unknown SID as a viewer as well.
        self.check_run(f"samba-tool service-account group-msa-membership add --name={gmsa_account} --principal={unknown_sid} -H {HOST} {ADMIN_CREDS}")

        # Read the SDDL again and check if the machine account and unknown SID were added.
        out = self.check_output(f"samba-tool service-account view --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")
        gmsa = json.loads(out.decode())
        machine_user = Computer.get(self.samdb, account_name=machine_account)
        expected_sddl = (initial_sddl +
                         f"(A;;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;{machine_user.object_sid})" +
                         f"(A;;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;{unknown_sid})")
        self.assertEqual(gmsa["msDS-GroupMSAMembership"], expected_sddl)

        # Get the list as --json which is easier to parse in the test.
        out = self.check_output(f"samba-tool service-account group-msa-membership show --name={gmsa_account} --json -H {HOST} {ADMIN_CREDS}")
        response = json.loads(out.decode())
        self.assertListEqual(response["trustees"], [
            "CN=Administrator,CN=Users,DC=addom,DC=samba,DC=example,DC=com",
            "CN=Machine_Account,CN=Computers,DC=addom,DC=samba,DC=example,DC=com",
            f"<SID={unknown_sid}>",
        ])

    def test_custom_sddl_complex(self):
        """Test custom SDDL that cannot be display as a simple list.

        In this case the "samba-tool service-account view" command
        can be used to retrieve the SDDL.
        """
        machine_account = "Machine_Account$"
        machine_password = "T3stPassword0nly"
        gmsa_account = "GMSA_Test_User$"

        # Create a machine account and set the password.
        self.check_run(f"samba-tool computer create {machine_account} -H {HOST} {ADMIN_CREDS}")
        self.addCleanup(self.run_command, f"samba-tool computer delete {machine_account} -H {HOST} {ADMIN_CREDS}")
        self.check_run(f"samba-tool user setpassword {machine_account} --newpassword={machine_password} -H {HOST} {ADMIN_CREDS}")

        # Create GMSA with custom SDDL this time rather than the command default.
        initial_sddl = f"O:SYD:(A;;RP;;;{self.samdb.connecting_user_sid})"
        self.check_run(f'samba-tool service-account create --name={gmsa_account} --dns-host-name=example.com --group-msa-membership="{initial_sddl}" --managed-password-interval=1 -H {HOST} {ADMIN_CREDS}')
        self.addCleanup(self.run_command, f"samba-tool service-account delete --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")

        # At first retrieving as a list will work fine.
        out = self.check_output(f"samba-tool service-account group-msa-membership show --name={gmsa_account} --json -H {HOST} {ADMIN_CREDS}")
        response = json.loads(out.decode())
        self.assertListEqual(
            response["trustees"],
            ["CN=Administrator,CN=Users,DC=addom,DC=samba,DC=example,DC=com"])

        # Set custom SDDL this time using the service-account modify command.
        machine_user = Computer.get(self.samdb, account_name=machine_account)
        deny_ace = f"(D;;CCDCLCSWRPWPDTLOCRSDRCWDWO;;;{machine_user.object_sid})"
        sddl = initial_sddl + deny_ace
        self.check_run(f'samba-tool service-account modify --name={gmsa_account} --group-msa-membership="{sddl}" -H {HOST} {ADMIN_CREDS}')

        # Group MSA membership can no longer be represented as a simple list.
        with self.assertRaisesRegex(BlackboxProcessError, "Cannot be represented as a simple list"):
            self.check_run(f"samba-tool service-account group-msa-membership show --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")

        # Retrieving the SDDL still works fine.
        out = self.check_output(f"samba-tool service-account view --name={gmsa_account} -H {HOST} {ADMIN_CREDS}")
        gmsa = json.loads(out.decode())
        self.assertEqual(gmsa["msDS-GroupMSAMembership"], sddl)


if __name__ == "__main__":
    import unittest
    unittest.main()



# GPO Parser for audit extensions
#
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2018
# Written by Garming Sam <garming@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import codecs
import csv
import io

from io import BytesIO
from xml.etree.ElementTree import Element, SubElement
from samba.gp_parse import GPParser
# [MS-GPAC] Group Policy Audit Configuration
class GPAuditCsvParser(GPParser):
    encoding = 'utf-8'
    header = None
    lines = []

    def parse(self, contents):
        self.lines = []
        reader = csv.reader(codecs.getreader(self.encoding)(BytesIO(contents)))

        self.header = next(reader)
        for row in reader:
            line = {}
            for i, x in enumerate(row):
                line[self.header[i]] = x

            self.lines.append(line)
            # print line

    def write_xml(self, filename):
        with open(filename, 'wb') as f:
            root = Element('CsvFile')
            child = SubElement(root, 'Row')
            for e in self.header:
                value = SubElement(child, 'Value')
                value.text = e

            for line in self.lines:
                child = SubElement(root, 'Row')
                for e, title in [(line[x], x) for x in self.header]:
                    value = SubElement(child, 'Value')
                    value.text = e

                    # Metadata for generalization
                    if title == 'Policy Target' and e != '':
                        value.attrib['user_id'] = 'TRUE'
                    if (title == 'Setting Value' and e != '' and
                        (line['Subcategory'] == 'RegistryGlobalSacl' or
                         line['Subcategory'] == 'FileGlobalSacl')):
                        value.attrib['acl'] = 'TRUE'

            self.write_pretty_xml(root, f)


        # contents = codecs.open(filename, encoding='utf-8').read()
        # self.load_xml(fromstring(contents))

    def load_xml(self, root):
        header = True
        self.lines = []

        for r in root.findall('Row'):
            if header:
                header = False
                self.header = []
                for v in r.findall('Value'):
                    if not isinstance(v.text, str):
                        v.text = v.text.decode(self.output_encoding)
                    self.header.append(v.text)
            else:
                line = {}
                for i, v in enumerate(r.findall('Value')):
                    line[self.header[i]] = v.text if v.text is not None else ''
                    if not isinstance(self.header[i], str):
                        line[self.header[i]] = line[self.header[i]].decode(self.output_encoding)

                self.lines.append(line)

    def write_binary(self, filename):
        from io import open
        with open(filename, 'w', encoding=self.encoding) as f:
            # In this case "binary" means "utf-8", so we let Python do that.
            writer = csv.writer(f, quoting=csv.QUOTE_MINIMAL)
            writer.writerow(self.header)
            for line in self.lines:
                writer.writerow([line[x] for x in self.header])



# Copyright (C) Volker Lendecke <vl@samba.org> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from samba.samba3.libsmb_samba_cwrapper import *
from samba.dcerpc import security

class Conn(LibsmbCConn):
    def deltree(self, path):
        if self.chkpath(path):
            for entry in self.list(path):
                self.deltree(path + "\\" + entry['name'])
            self.rmdir(path)
        else:
            self.unlink(path)

    SECINFO_DEFAULT_FLAGS = \
        security.SECINFO_OWNER | \
        security.SECINFO_GROUP | \
        security.SECINFO_DACL | \
        security.SECINFO_SACL

    def required_access_for_get_secinfo(self, secinfo):
        access = 0

        #
        # This is based on MS-FSA
        # 2.1.5.13 Server Requests a Query of Security Information
        #
        # Note that MS-SMB2 3.3.5.20.3 Handling SMB2_0_INFO_SECURITY
        # doesn't specify any extra checks
        #

        if secinfo & security.SECINFO_OWNER:
            access |= security.SEC_STD_READ_CONTROL
        if secinfo & security.SECINFO_GROUP:
            access |= security.SEC_STD_READ_CONTROL
        if secinfo & security.SECINFO_DACL:
            access |= security.SEC_STD_READ_CONTROL
        if secinfo & security.SECINFO_SACL:
            access |= security.SEC_FLAG_SYSTEM_SECURITY

        if secinfo & security.SECINFO_LABEL:
            access |= security.SEC_STD_READ_CONTROL

        return access

    def required_access_for_set_secinfo(self, secinfo):
        access = 0

        #
        # This is based on MS-FSA
        # 2.1.5.16 Server Requests Setting of Security Information
        # and additional constraints from
        # MS-SMB2 3.3.5.21.3 Handling SMB2_0_INFO_SECURITY
        #

        if secinfo & security.SECINFO_OWNER:
            access |= security.SEC_STD_WRITE_OWNER
        if secinfo & security.SECINFO_GROUP:
            access |= security.SEC_STD_WRITE_OWNER
        if secinfo & security.SECINFO_DACL:
            access |= security.SEC_STD_WRITE_DAC
        if secinfo & security.SECINFO_SACL:
            access |= security.SEC_FLAG_SYSTEM_SECURITY

        if secinfo & security.SECINFO_LABEL:
            access |= security.SEC_STD_WRITE_OWNER

        if secinfo & security.SECINFO_ATTRIBUTE:
            access |= security.SEC_STD_WRITE_DAC

        if secinfo & security.SECINFO_SCOPE:
            access |= security.SEC_FLAG_SYSTEM_SECURITY

        if secinfo & security.SECINFO_BACKUP:
            access |= security.SEC_STD_WRITE_OWNER
            access |= security.SEC_STD_WRITE_DAC
            access |= security.SEC_FLAG_SYSTEM_SECURITY

        return access

    def get_acl(self,
                filename,
                sinfo=None,
                access_mask=None):
        """Get security descriptor for file."""
        if sinfo is None:
            sinfo = self.SECINFO_DEFAULT_FLAGS
        if access_mask is None:
            access_mask = self.required_access_for_get_secinfo(sinfo)
        fnum = self.create(
            Name=filename,
            DesiredAccess=access_mask,
            ShareAccess=(FILE_SHARE_READ|FILE_SHARE_WRITE))
        try:
            sd = self.get_sd(fnum, sinfo)
        finally:
            self.close(fnum)
        return sd

    def set_acl(self,
                filename,
                sd,
                sinfo=None,
                access_mask=None):
        """Set security descriptor for file."""
        if sinfo is None:
            sinfo = self.SECINFO_DEFAULT_FLAGS
        if access_mask is None:
            access_mask = self.required_access_for_set_secinfo(sinfo)
        fnum = self.create(
            Name=filename,
            DesiredAccess=access_mask,
            ShareAccess=(FILE_SHARE_READ|FILE_SHARE_WRITE))
        try:
            self.set_sd(fnum, sd, sinfo)
        finally:
            self.close(fnum)



# user management
#
# disable user
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba import ldb
from samba.auth import system_session
from samba.netcmd import Command, CommandError, Option
from samba.samdb import SamDB
from samba.dcerpc import security
from samba.ndr import ndr_unpack


class cmd_user_disable(Command):
    """Disable a user."""

    synopsis = "%prog (<username>|--filter <filter>) [options]"

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        Option("--filter",
               help="LDAP filter to select user",
               type=str,
               dest="search_filter"),
        Option("--remove-supplemental-groups",
               help="Remove user's supplemental groups",
               action="store_true"),
    ]

    takes_args = ["username?"]

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "versionopts": options.VersionOptions,
    }

    def run(self, username=None, sambaopts=None, credopts=None,
            versionopts=None, search_filter=None, H=None,
            remove_supplemental_groups=False):
        if username is None and search_filter is None:
            raise CommandError("Either the username or '--filter' must be specified!")

        if search_filter is None:
            search_filter = "(&(objectClass=user)(sAMAccountName=%s))" % (
                ldb.binary_encode(username))

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp, fallback_machine=True)

        samdb = SamDB(url=H, session_info=system_session(),
                      credentials=creds, lp=lp)

        samdb.transaction_start()
        try:
            res = samdb.search(base=samdb.domain_dn(),
                               expression=search_filter,
                               scope=ldb.SCOPE_SUBTREE,
                               controls=["extended_dn:1:1"],
                               attrs=["objectSid", "memberOf"])
            user_groups = res[0].get("memberOf")
            if user_groups is None:
                user_groups = []
            user_binary_sid = res[0].get("objectSid", idx=0)
            user_sid = ndr_unpack(security.dom_sid, user_binary_sid)
        except IndexError:
            samdb.transaction_cancel()
            raise CommandError("Unable to find user '%s'" % (
                               username or search_filter))
        except Exception as msg:
            samdb.transaction_cancel()
            raise CommandError("Failed to find user '%s': '%s'" % (
                               username or search_filter, msg))
        if len(res) > 1:
            samdb.transaction_cancel()
            raise CommandError("Found more than one user '%s'" % (
                               username or search_filter))

        if remove_supplemental_groups:
            for user_group in user_groups:
                try:
                    samdb.add_remove_group_members(str(user_group),
                                                   [str(user_sid)],
                                                   add_members_operation=False)
                except Exception as msg:
                    samdb.transaction_cancel()
                    raise CommandError("Failed to remove user from group "
                                       "'%s': %s" % (user_group, msg))

        try:
            samdb.disable_account(search_filter)
        except Exception as msg:
            samdb.transaction_cancel()
            raise CommandError("Failed to disable user '%s': %s" % (
                username or search_filter, msg))
        samdb.transaction_commit()
